import malaya

def test_basic_stemmer():
    assert malaya.basic_normalizer('x masing2') == 'tidak masing-masing'

def test_naive_stemmer():
    corpus_normalize = ['maka','serious','yeke','masing-masing']
    normalizer = malaya.naive_normalizer(corpus_normalize)
    assert normalizer.normalize('masing2') == 'masing-masing'
    assert normalizer.normalize('xmasing2') == 'tak masing-masing'
    assert normalizer.normalize('x') == 'tak'



from malaya.text_functions import malaya_textcleaning

def test_malaya_textcleaning():
    assert len(malaya_textcleaning('saya sebenarnya sukakan awak, hahahaha')) > 0

def test_malaya_textcleaning_trash():
    assert not len(malaya_textcleaning('asdsad asdsad easdcxv'))



import malaya

def test_word2vec_n_closest():
    embedded = malaya.malaya_word2vec(256)
    word_vector = malaya.Word2Vec(embedded['nce_weights'], embedded['dictionary'])
    word = 'anwar'
    assert len(word_vector.n_closest(word=word, num_closest=8, metric='cosine'))

def test_word2vec_n_closest_without_similarity():
    embedded = malaya.malaya_word2vec(256)
    word_vector = malaya.Word2Vec(embedded['nce_weights'], embedded['dictionary'])
    word = 'anwar'
    assert len(word_vector.n_closest(word=word, num_closest=8, metric='cosine', return_similarity=False))

def test_word2vec_calculator_n_closest():
    embedded = malaya.malaya_word2vec(256)
    word_vector = malaya.Word2Vec(embedded['nce_weights'], embedded['dictionary'])
    assert len(word_vector.calculator('anwar + mahathir', num_closest=8, metric='cosine'))

def test_word2vec_calculator_minus():
    embedded = malaya.malaya_word2vec(256)
    word_vector = malaya.Word2Vec(embedded['nce_weights'], embedded['dictionary'])
    assert len(word_vector.calculator('anwar - mahathir', num_closest=8, metric='cosine'))

def test_word2vec_calculator_multiply():
    embedded = malaya.malaya_word2vec(256)
    word_vector = malaya.Word2Vec(embedded['nce_weights'], embedded['dictionary'])
    assert len(word_vector.calculator('anwar * mahathir', num_closest=8, metric='cosine'))

def test_word2vec_calculator_divide():
    embedded = malaya.malaya_word2vec(256)
    word_vector = malaya.Word2Vec(embedded['nce_weights'], embedded['dictionary'])
    assert len(word_vector.calculator('anwar / mahathir', num_closest=8, metric='cosine'))

def test_word2vec_calculator_bracket():
    embedded = malaya.malaya_word2vec(256)
    word_vector = malaya.Word2Vec(embedded['nce_weights'], embedded['dictionary'])
    assert len(word_vector.calculator('(anwar+hadi) * mahathir', num_closest=8, metric='cosine'))

def test_word2vec_calculator_n_closest_without_similarity():
    embedded = malaya.malaya_word2vec(256)
    word_vector = malaya.Word2Vec(embedded['nce_weights'], embedded['dictionary'])
    assert len(word_vector.calculator('anwar + mahathir', num_closest=8, metric='cosine', return_similarity=False))

def test_word2vec_analogy():
    embedded = malaya.malaya_word2vec(256)
    word_vector = malaya.Word2Vec(embedded['nce_weights'], embedded['dictionary'])
    assert len(word_vector.analogy('anwar', 'penjara', 'kerajaan', 5)) == 5

def test_word2vec_tsne():
    embedded = malaya.malaya_word2vec(32)
    word_vector = malaya.Word2Vec(embedded['nce_weights'], embedded['dictionary'])
    embed_2d, word_list = word_vector.project_2d(0, 100)
    assert embed_2d.shape[1] == 2



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from itertools import product
from collections import defaultdict
import numpy as np
from scipy.spatial.distance import euclidean
import pulp


def _tokens_to_fracdict(tokens):
    cntdict = defaultdict(lambda: 0)
    for token in tokens:
        cntdict[token] += 1
    totalcnt = sum(cntdict.values())
    return {token: float(cnt) / totalcnt for token, cnt in cntdict.items()}


def _word_mover(left_token, right_token, vectorizer):
    all_tokens = list(set(left_token + right_token))
    wordvecs = {
        token: vectorizer.get_vector_by_name(token) for token in all_tokens
    }
    left_bucket = _tokens_to_fracdict(left_token)
    right_bucket = _tokens_to_fracdict(right_token)

    T = pulp.LpVariable.dicts(
        'T_matrix', list(product(all_tokens, all_tokens)), lowBound = 0
    )
    prob = pulp.LpProblem('WMD', sense = pulp.LpMinimize)
    prob += pulp.lpSum(
        [
            T[token1, token2] * euclidean(wordvecs[token1], wordvecs[token2])
            for token1, token2 in product(all_tokens, all_tokens)
        ]
    )
    for token2 in right_bucket:
        prob += (
            pulp.lpSum([T[token1, token2] for token1 in left_bucket])
            == right_bucket[token2]
        )
    for token1 in left_bucket:
        prob += (
            pulp.lpSum([T[token1, token2] for token2 in right_bucket])
            == left_bucket[token1]
        )
    prob.solve()
    return prob


def distance(left_token, right_token, vectorizer):
    """
    calculate word mover distance between left hand-side sentence and right hand-side sentence

    Parameters
    ----------
    left_token : list
        Eg, ['saya','suka','makan','ayam']
    right_token : list
        Eg, ['saya','suka','makan','ikan']
    vectorizer : object
        fast-text or word2vec interface object

    Returns
    -------
    distance: float
    """
    assert isinstance(left_token, list), 'left_token must be a list'
    assert isinstance(right_token, list), 'right_token must be a list'
    assert hasattr(
        vectorizer, 'get_vector_by_name'
    ), 'vectorizer must has `get_vector_by_name` method'
    prob = _word_mover(left_token, right_token, vectorizer)
    return pulp.value(prob.objective)



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import pickle
import json
import tensorflow as tf
from ._utils._utils import check_file, load_graph
from ._models._sklearn_model import DEPENDENCY
from ._models._tensorflow_model import DEPENDENCY as TF_DEPENDENCY
from ._utils._parse_dependency import DependencyGraph
from ._utils._paths import PATH_DEPEND, S3_PATH_DEPEND


def dependency_graph(tagging, indexing):
    """
    Return helper object for dependency parser results. Only accept tagging and indexing outputs from dependency models
    """
    result = []
    for i in range(len(tagging)):
        result.append(
            '%d\t%s\t_\t_\t_\t_\t%d\t%s\t_\t_'
            % (i + 1, tagging[i][0], int(indexing[i][1]), tagging[i][1])
        )
    return DependencyGraph('\n'.join(result), top_relation_label = 'root')


def available_deep_model():
    """
    List available deep learning dependency models, ['concat', 'bahdanau', 'luong']
    """
    return ['concat', 'bahdanau', 'luong']


def crf(validate = True):
    """
    Load CRF dependency model.

    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    DEPENDENCY : malaya._models._sklearn_model.DEPENDENCY class
    """
    if validate:
        check_file(PATH_DEPEND['crf'], S3_PATH_DEPEND['crf'])
    else:
        if not check_available(PATH_DEPEND['crf']):
            raise Exception(
                'dependency/crf is not available, please `validate = True`'
            )
    try:
        with open(PATH_DEPEND['crf']['model'], 'rb') as fopen:
            model = pickle.load(fopen)
        with open(PATH_DEPEND['crf']['depend'], 'rb') as fopen:
            depend = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('dependency/crf') and try again"
        )
    return DEPENDENCY(model, depend)


def deep_model(model = 'bahdanau', validate = True):
    """
    Load deep learning dependency model.

    Parameters
    ----------
    model : str, optional (default='bahdanau')
        Model architecture supported. Allowed values:

        * ``'concat'`` - Concating character and word embedded for BiLSTM
        * ``'bahdanau'`` - Concating character and word embedded including Bahdanau Attention for BiLSTM
        * ``'luong'`` - Concating character and word embedded including Luong Attention for BiLSTM
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    DEPENDENCY: malaya._models._tensorflow_model.DEPENDENCY class
    """
    assert isinstance(model, str), 'model must be a string'
    model = model.lower()
    if model in ['concat', 'bahdanau', 'luong']:
        if validate:
            check_file(PATH_DEPEND[model], S3_PATH_DEPEND[model])
        else:
            if not check_available(PATH_DEPEND[model]):
                raise Exception(
                    'dependency/%s is not available, please `validate = True`'
                    % (model)
                )
        try:
            with open(PATH_DEPEND[model]['setting'], 'r') as fopen:
                nodes = json.loads(fopen.read())
            g = load_graph(PATH_DEPEND[model]['model'])
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('dependency/%s') and try again"
                % (model)
            )
        return TF_DEPENDENCY(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/Placeholder_1:0'),
            g.get_tensor_by_name('import/logits:0'),
            g.get_tensor_by_name('import/logits_depends:0'),
            nodes,
            tf.InteractiveSession(graph = g),
            model,
            g.get_tensor_by_name('import/transitions:0'),
            g.get_tensor_by_name('import/depends/transitions:0'),
            g.get_tensor_by_name('import/Variable:0'),
        )

    else:
        raise Exception(
            'model not supported, please check supported models from malaya.dependency.available_deep_model()'
        )



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import re
import os
import random
import numpy as np
from fuzzywuzzy import fuzz
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.utils import shuffle
from .texts._text_functions import (
    STOPWORDS,
    STOPWORD_CALON,
    simple_textcleaning,
    str_idx,
)
from .cluster import sentence_ngram
from . import home, _delete_macos
from ._utils._utils import download_file
from ._models._skip_thought import train_model as skip_train, batch_sequence
from ._models._siamese_lstm import train_model as siamese_train

_zip_location = home + '/rules-based.zip'
_namacalon = None
_parlimen = None
_dun = None
_negeri = None
_location = None
_person_dict = None
_topic_dict = None
_short_dict = None


def _apply_stopwords_calon(string):
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string.lower()).strip()
    return ' '.join(
        [i for i in string.split() if i not in STOPWORD_CALON and len(i) > 1]
    )


class _DEEP_SIAMESE_SIMILARITY:
    def __init__(
        self, sess, model, keys, dictionary, maxlen, is_influencers = False
    ):
        self._sess = sess
        self._model = model
        self.keys = keys
        self.dictionary = dictionary
        self.maxlen = maxlen
        self._is_influencers = is_influencers

    def get_similarity(self, string, anchor = 0.5):
        """
        Return similar topics / influencers.

        Parameters
        ----------
        string: str
        anchor: float, (default=0.5)
            baseline similarity.

        Returns
        -------
        results: list of strings
        """
        assert isinstance(string, str), 'input must be a string'
        assert isinstance(anchor, float), 'anchor must be a float'
        assert (
            anchor > 0 and anchor < 1
        ), 'anchor must be bigger than 0, less than 1'
        original_string = simple_textcleaning(string)
        strings = [original_string] * len(self.keys)
        left = str_idx(strings, self.dictionary, self.maxlen, UNK = 3)
        right = str_idx(self.keys, self.dictionary, self.maxlen, UNK = 3)
        distances = self._sess.run(
            1 - self._model.distance,
            feed_dict = {self._model.X_left: left, self._model.X_right: right},
        )
        where = np.where(distances > anchor)[0]
        results = [self.keys[i].lower() for i in where]
        for key, vals in _short_dict.items():
            for v in vals:
                if v in original_string.split():
                    results.append(key.lower())
                    break
        if self._is_influencers:
            for index in np.where(
                np.array(
                    [
                        fuzz.token_set_ratio(i, original_string)
                        for i in _namacalon
                    ]
                )
                >= 80
            )[0]:
                results.append(_namacalon[index].lower())
        return list(set(results))


class _DEEP_SIMILARITY:
    def __init__(
        self,
        sess,
        model,
        vectorized,
        keys,
        dictionary,
        maxlen,
        is_influencers = False,
    ):
        self._sess = sess
        self._model = model
        self.vectorized = vectorized
        self.keys = keys
        self.dictionary = dictionary
        self.maxlen = maxlen
        self._is_influencers = is_influencers

    def get_similarity(self, string, anchor = 0.5):
        """
        Return similar topics / influencers.

        Parameters
        ----------
        string: str
        anchor: float, (default=0.5)
            baseline similarity.

        Returns
        -------
        results: list of strings
        """
        assert isinstance(string, str), 'input must be a string'
        assert isinstance(anchor, float), 'anchor must be a float'
        assert (
            anchor > 0 and anchor < 1
        ), 'anchor must be bigger than 0, less than 1'
        original_string = simple_textcleaning(string)
        string = ' '.join(set(original_string.split()))
        encoded = self._sess.run(
            self._model.get_thought,
            feed_dict = {
                self._model.INPUT: batch_sequence(
                    [string], self.dictionary, maxlen = self.maxlen
                )
            },
        )
        where = np.where(
            cosine_similarity(self.vectorized, encoded)[:, 0] > anchor
        )[0]
        results = [self.keys[i].lower() for i in where]
        for key, vals in _short_dict.items():
            for v in vals:
                if v in original_string.split():
                    results.append(key.lower())
                    break
        if self._is_influencers:
            for index in np.where(
                np.array(
                    [
                        fuzz.token_set_ratio(i, original_string)
                        for i in _namacalon
                    ]
                )
                >= 80
            )[0]:
                results.append(_namacalon[index].lower())
        return list(set(results))


class _FAST_SIMILARITY:
    def __init__(self, vectorizer, vectorized, keys, is_influencers = False):
        self.vectorizer = vectorizer
        self.vectorized = vectorized
        self.keys = keys
        self._is_influencers = is_influencers

    def get_similarity(self, string, anchor = 0.1):
        """
        Return similar topics / influencers.

        Parameters
        ----------
        string: str
        anchor: float, (default=0.5)
            baseline similarity.

        Returns
        -------
        results: list of strings
        """
        assert isinstance(string, str), 'input must be a string'
        assert isinstance(anchor, float), 'anchor must be a float'
        assert (
            anchor > 0 and anchor < 1
        ), 'anchor must be bigger than 0, less than 1'
        original_string = simple_textcleaning(string)
        string = ' '.join(set(original_string.split()))
        where = np.where(
            cosine_similarity(
                self.vectorized, self.vectorizer.transform([string])
            )[:, 0]
            > anchor
        )[0]
        results = [self.keys[i].lower() for i in where]
        for key, vals in _short_dict.items():
            for v in vals:
                if v in original_string.split():
                    results.append(key.lower())
                    break
        if self._is_influencers:
            for index in np.where(
                np.array(
                    [
                        fuzz.token_set_ratio(i, original_string)
                        for i in _namacalon
                    ]
                )
                >= 80
            )[0]:
                results.append(_namacalon[index].lower())
        return list(set(results))


def _load_internal_data():
    import pandas as pd
    import zipfile

    global _namacalon, _parlimen, _dun, _negeri, _location, _person_dict, _topic_dict, _short_dict
    if not os.path.isfile(_zip_location):
        print('downloading Topics, Influencers, location data')
        download_file('rules-based.zip', _zip_location)

    if not os.path.exists(home + '/rules-based/calon.csv'):
        with zipfile.ZipFile(_zip_location, 'r') as zip_ref:
            zip_ref.extractall(home)

    _delete_macos()
    df = pd.read_csv(home + '/rules-based/calon.csv')
    _namacalon = df.NamaCalon.str.lower().unique().tolist()
    for i in range(len(_namacalon)):
        _namacalon[i] = _apply_stopwords_calon(_namacalon[i])
    _namacalon = list(set(_namacalon))

    df = pd.read_csv(home + '/rules-based/negeri.csv')
    _negeri = df.negeri.str.lower().unique().tolist()
    _parlimen = df.parlimen.str.lower().unique().tolist()
    _dun = df.dun.str.lower().unique().tolist()[:-1]

    _location = _negeri + _parlimen + _dun

    with open(home + '/rules-based/person-normalized', 'r') as fopen:
        person = list(filter(None, fopen.read().split('\n')))

    _person_dict = {}
    for i in range(len(person)):
        splitted = person[i].split(':')
        uniques = list(
            filter(
                None,
                (
                    set(
                        [k.strip().lower() for k in splitted[1].split(', ')]
                        + [splitted[0].lower()]
                    )
                ),
            )
        )
        _person_dict[splitted[0]] = uniques

    with open(home + '/rules-based/topic-normalized', 'r') as fopen:
        topic = list(filter(None, fopen.read().split('\n')))

    _topic_dict = {}
    for i in range(len(topic)):
        splitted = topic[i].split(':')
        uniques = list(
            filter(
                None,
                (
                    set(
                        [k.strip().lower() for k in splitted[1].split(', ')]
                        + [splitted[0].lower()]
                    )
                ),
            )
        )
        _topic_dict[splitted[0]] = uniques

    with open(home + '/rules-based/short-normalized', 'r') as fopen:
        short = list(filter(None, fopen.read().split('\n')))

    _short_dict = {}
    for i in range(len(short)):
        splitted = short[i].split(':')
        uniques = list(
            filter(
                None,
                (
                    set(
                        [k.strip().lower() for k in splitted[1].split(', ')]
                        + [splitted[0].lower()]
                    )
                ),
            )
        )
        _short_dict[splitted[0]] = uniques


def fuzzy_influencer(string):
    """
    Return similar influencers.

    Parameters
    ----------
    string: str

    Returns
    -------
    results: list of strings
    """
    if not _person_dict:
        _load_internal_data()
    assert isinstance(string, str), 'input must be a string'
    string = string.lower()
    influencers = []
    for key, vals in _person_dict.items():
        for v in vals:
            if fuzz.token_set_ratio(v, string) >= 80:
                influencers.append(key.lower())
                break
    for key, vals in _short_dict.items():
        for v in vals:
            if v in string.split():
                influencers.append(key.lower())
                break

    for index in np.where(
        np.array([fuzz.token_set_ratio(i, string) for i in _namacalon]) >= 80
    )[0]:
        influencers.append(_namacalon[index].lower())
    return list(set(influencers))


def fuzzy_topic(string):
    """
    Return similar topics.

    Parameters
    ----------
    string: str

    Returns
    -------
    results: list of strings
    """
    if not _person_dict:
        _load_internal_data()
    assert isinstance(string, str), 'input must be a string'
    string = string.lower()
    topics = []
    for key, vals in _topic_dict.items():
        for v in vals:
            if fuzz.token_set_ratio(v, string) >= 80:
                topics.append(key.lower())
                break
    for key, vals in _person_dict.items():
        for v in vals:
            if fuzz.token_set_ratio(v, string) >= 80:
                topics.append(key.lower())
                break
    for key, vals in _short_dict.items():
        for v in vals:
            if v in string.split():
                topics.append(key.lower())
                break

    return list(set(topics))


def is_location(string):
    """
    check whether a string is a malaysia location.

    Parameters
    ----------
    string: str

    Returns
    -------
    boolean: bool
    """
    if not _person_dict:
        _load_internal_data()
    for loc in _location:
        if fuzz.token_set_ratio(loc.lower(), string) >= 90:
            return True
    return False


def fuzzy_location(string):
    """
    Return similar _location.

    Parameters
    ----------
    string: str

    Returns
    -------
    results: list of strings
    """
    if not _person_dict:
        _load_internal_data()
    assert isinstance(string, str), 'input must be a string'
    negeri_list = list(
        set([i for i in _negeri if fuzz.token_set_ratio(i, string) >= 80])
    )
    parlimen_list = list(
        set([i for i in _parlimen if fuzz.token_set_ratio(i, string) >= 80])
    )
    dun_list = list(
        set([i for i in _dun if fuzz.token_set_ratio(i, string) >= 80])
    )
    return {'negeri': negeri_list, 'parlimen': parlimen_list, 'dun': dun_list}


def _generate_topics():
    if not _person_dict:
        _load_internal_data()
    texts = [' '.join(words) for _, words in _topic_dict.items()]
    keys = [key for key, _ in _topic_dict.items()]
    texts += [' '.join(words) for _, words in _person_dict.items()]
    keys += [key for key, _ in _person_dict.items()]
    texts = [' '.join(list(set(text.split()))) for text in texts]
    output = []
    for text in texts:
        output.append(
            ' '.join([word for word in text.split() if word not in STOPWORDS])
        )
    return output, keys


def _generate_influencers():
    if not _person_dict:
        _load_internal_data()
    texts = [' '.join(words) for _, words in _person_dict.items()]
    keys = [key for key, _ in _person_dict.items()]
    texts = [' '.join(list(set(text.split()))) for text in texts]
    output = []
    for text in texts:
        output.append(
            ' '.join([word for word in text.split() if word not in STOPWORDS])
        )
    return output, keys


def siamese_topic(
    epoch = 5,
    batch_size = 32,
    embedding_size = 256,
    output_size = 100,
    maxlen = 100,
    ngram = (1, 4),
):
    """
    Train a deep siamese network for topics similarity

    Parameters
    ----------
    epoch: int, (default=5)
        iteration numbers
    batch_size: int, (default=32)
        batch size for every feed, batch size must <= size of corpus
    embedding_size: int, (default=256)
        vector size representation for a word
    output_size: int, (default=100)
        encoder output size, bigger means more vector definition
    maxlen: int, (default=100)
        max length of a string to be train
    ngram: tuple, (default=(1,4))
        n-grams size to train a corpus

    Returns
    -------
    _DEEP_SIAMESE_SIMILARITY: malaya.topics_influencers._DEEP_SIAMESE_SIMILARITY class
    """
    if not _person_dict:
        _load_internal_data()
    assert isinstance(epoch, int), 'epoch must be an integer'
    assert isinstance(batch_size, int), 'batch_size must be an integer'
    assert isinstance(embedding_size, int), 'embedding_size must be an integer'
    assert isinstance(output_size, int), 'output_size must be an integer'
    assert isinstance(maxlen, int), 'maxlen must be an integer'
    assert isinstance(ngram, tuple), 'ngram must be a tuple'
    assert len(ngram) == 2, 'ngram size must equal to 2'
    output, keys = _generate_topics()
    batch_x_left, batch_x_right, batch_y = [], [], []
    for i in range(len(output)):
        augmentation = sentence_ngram(output[i])
        batch_x_right.extend([keys[i]] * len(augmentation))
        batch_x_left.extend(augmentation)
        batch_y.extend([1] * len(augmentation))
    neg_batch_x_left, neg_batch_x_right = [], []
    for i in range(len(batch_x_left)):
        random_num = random.randint(0, len(batch_x_left) - 1)
        while random_num == i:
            random_num = random.randint(0, len(batch_x_left) - 1)
        neg_batch_x_right.append(batch_x_right[random_num])
        neg_batch_x_left.append(batch_x_left[i])
        batch_y.append(0)
    batch_x_left.extend(neg_batch_x_left)
    batch_x_right.extend(neg_batch_x_right)
    batch_x_left, batch_x_right, batch_y = shuffle(
        batch_x_left, batch_x_right, batch_y
    )
    sess, model, dictionary = siamese_train(
        batch_x_left,
        batch_x_right,
        batch_y,
        epoch = epoch,
        batch_size = batch_size,
        embedding_size = embedding_size,
        output_size = output_size,
        maxlen = maxlen,
    )
    return _DEEP_SIAMESE_SIMILARITY(sess, model, keys, dictionary, maxlen)


def siamese_influencer(
    epoch = 5,
    batch_size = 32,
    embedding_size = 256,
    output_size = 100,
    maxlen = 100,
    ngram = (1, 4),
):
    """
    Train a deep siamese network for influencers similarity

    Parameters
    ----------
    epoch: int, (default=5)
        iteration numbers
    batch_size: int, (default=32)
        batch size for every feed, batch size must <= size of corpus
    embedding_size: int, (default=256)
        vector size representation for a word
    output_size: int, (default=100)
        encoder output size, bigger means more vector definition
    maxlen: int, (default=100)
        max length of a string to be train
    ngram: tuple, (default=(1,4))
        n-grams size to train a corpus

    Returns
    -------
    _DEEP_SIAMESE_SIMILARITY: malaya.topic_influencers._DEEP_SIAMESE_SIMILARITY class
    """
    if not _person_dict:
        _load_internal_data()
    assert isinstance(epoch, int), 'epoch must be an integer'
    assert isinstance(batch_size, int), 'batch_size must be an integer'
    assert isinstance(embedding_size, int), 'embedding_size must be an integer'
    assert isinstance(output_size, int), 'output_size must be an integer'
    assert isinstance(maxlen, int), 'maxlen must be an integer'
    assert isinstance(ngram, tuple), 'ngram must be a tuple'
    assert len(ngram) == 2, 'ngram size must equal to 2'
    output, keys = _generate_influencers()
    batch_x_left, batch_x_right, batch_y = [], [], []
    for i in range(len(output)):
        augmentation = sentence_ngram(output[i])
        batch_x_right.extend([keys[i]] * len(augmentation))
        batch_x_left.extend(augmentation)
        batch_y.extend([1] * len(augmentation))
    neg_batch_x_left, neg_batch_x_right = [], []
    for i in range(len(batch_x_left)):
        random_num = random.randint(0, len(batch_x_left) - 1)
        while random_num == i:
            random_num = random.randint(0, len(batch_x_left) - 1)
        neg_batch_x_right.append(batch_x_right[random_num])
        neg_batch_x_left.append(batch_x_left[i])
        batch_y.append(0)
    batch_x_left.extend(neg_batch_x_left)
    batch_x_right.extend(neg_batch_x_right)
    batch_x_left, batch_x_right, batch_y = shuffle(
        batch_x_left, batch_x_right, batch_y
    )
    sess, model, dictionary = siamese_train(
        batch_x_left,
        batch_x_right,
        batch_y,
        epoch = epoch,
        batch_size = batch_size,
        embedding_size = embedding_size,
        output_size = output_size,
        maxlen = maxlen,
    )
    return _DEEP_SIAMESE_SIMILARITY(
        sess, model, keys, dictionary, maxlen, is_influencers = True
    )


def skipthought_topic(
    epoch = 5,
    batch_size = 16,
    embedding_size = 256,
    maxlen = 100,
    ngram = (1, 4),
):
    """
    Train a deep skip-thought network for topics similarity

    Parameters
    ----------
    epoch: int, (default=5)
        iteration numbers
    batch_size: int, (default=32)
        batch size for every feed, batch size must <= size of corpus
    embedding_size: int, (default=256)
        vector size representation for a word
    maxlen: int, (default=100)
        max length of a string to be train
    ngram: tuple, (default=(1,4))
        n-grams size to train a corpus

    Returns
    -------
    _DEEP_SIMILARITY: malaya.topic_influencers._DEEP_SIMILARITY class
    """
    if not _person_dict:
        _load_internal_data()
    assert isinstance(epoch, int), 'epoch must be an integer'
    assert isinstance(batch_size, int), 'batch_size must be an integer'
    assert isinstance(embedding_size, int), 'embedding_size must be an integer'
    assert isinstance(maxlen, int), 'maxlen must be an integer'
    assert isinstance(ngram, tuple), 'ngram must be a tuple'
    assert len(ngram) == 2, 'ngram size must equal to 2'
    output, keys = _generate_topics()
    batch_x, batch_y = [], []
    for i in range(len(output)):
        augmentation = sentence_ngram(output[i])
        batch_y.extend([keys[i]] * len(augmentation))
        batch_x.extend(augmentation)
    batch_x, batch_y = shuffle(batch_x, batch_y)
    sess, model, dictionary = skip_train(
        batch_x,
        batch_y,
        batch_y,
        epoch = epoch,
        batch_size = batch_size,
        embedding_size = embedding_size,
        maxlen = maxlen,
    )
    encoded = sess.run(
        model.get_thought,
        feed_dict = {
            model.INPUT: batch_sequence(output, dictionary, maxlen = maxlen)
        },
    )
    return _DEEP_SIMILARITY(sess, model, encoded, keys, dictionary, maxlen)


def skipthought_influencer(
    epoch = 10,
    batch_size = 16,
    embedding_size = 256,
    maxlen = 100,
    ngram = (1, 4),
):
    """
    Train a deep skip-thought network for influencers similarity

    Parameters
    ----------
    epoch: int, (default=5)
        iteration numbers
    batch_size: int, (default=32)
        batch size for every feed, batch size must <= size of corpus
    embedding_size: int, (default=256)
        vector size representation for a word
    maxlen: int, (default=100)
        max length of a string to be train
    ngram: tuple, (default=(1,4))
        n-grams size to train a corpus

    Returns
    -------
    _DEEP_SIMILARITY: malaya.topics_influencers._DEEP_SIMILARITY class
    """
    if not _person_dict:
        _load_internal_data()
    assert isinstance(epoch, int), 'epoch must be an integer'
    assert isinstance(batch_size, int), 'batch_size must be an integer'
    assert isinstance(embedding_size, int), 'embedding_size must be an integer'
    assert isinstance(maxlen, int), 'maxlen must be an integer'
    assert isinstance(ngram, tuple), 'ngram must be a tuple'
    assert len(ngram) == 2, 'ngram size must equal to 2'
    output, keys = _generate_influencers()
    batch_x, batch_y = [], []
    for i in range(len(output)):
        augmentation = sentence_ngram(output[i])
        batch_y.extend([keys[i]] * len(augmentation))
        batch_x.extend(augmentation)
    assert batch_size < len(batch_x), 'batch size must smaller with corpus size'
    batch_x, batch_y = shuffle(batch_x, batch_y)
    sess, model, dictionary = skip_train(
        batch_x,
        batch_y,
        batch_y,
        epoch = epoch,
        batch_size = batch_size,
        embedding_size = embedding_size,
        maxlen = maxlen,
    )
    encoded = sess.run(
        model.get_thought,
        feed_dict = {
            model.INPUT: batch_sequence(output, dictionary, maxlen = maxlen)
        },
    )
    return _DEEP_SIMILARITY(
        sess, model, encoded, keys, dictionary, maxlen, is_influencers = True
    )


def fast_topic(vectorizer = 'tfidf', ngram = (3, 10)):
    """
    Train a deep siamese network for topics similarity

    Parameters
    ----------
    vectorizer: str, (default='tfidf')
        vectorization technique for a corpus
    ngram: tuple, (default=(1,4))
        n-grams size to train a corpus

    Returns
    -------
    _FAST_SIMILARITY: malaya.siamese_lstm._FAST_SIMILARITY class
    """
    if not _person_dict:
        _load_internal_data()
    assert isinstance(vectorizer, str), 'vectorizer must be a string'
    assert isinstance(ngram, tuple), 'ngram must be a tuple'
    assert len(ngram) == 2, 'ngram size must equal to 2'
    if 'tfidf' in vectorizer.lower():
        char_vectorizer = TfidfVectorizer(
            sublinear_tf = True,
            strip_accents = 'unicode',
            analyzer = 'char',
            ngram_range = ngram,
        )
    elif 'count' in vectorizer.lower():
        char_vectorizer = CountVectorizer(
            strip_accents = 'unicode', analyzer = 'char', ngram_range = ngram
        )
    else:
        raise Exception('model not supported')
    output, keys = _generate_topics()
    vectorizer = char_vectorizer.fit(output)
    vectorized = vectorizer.transform(output)
    return _FAST_SIMILARITY(vectorizer, vectorized, keys)


def fast_influencer(vectorizer = 'tfidf', ngram = (3, 10)):
    """
    Train a deep siamese network for influencers similarity

    Parameters
    ----------
    vectorizer: str, (default='tfidf')
        vectorization technique for a corpus
    ngram: tuple, (default=(1,4))
        n-grams size to train a corpus

    Returns
    -------
    _FAST_SIMILARITY: malaya.siamese_lstm._FAST_SIMILARITY class
    """
    if not _person_dict:
        _load_internal_data()
    assert isinstance(vectorizer, str), 'vectorizer must be a string'
    assert isinstance(ngram, tuple), 'ngram must be a tuple'
    assert len(ngram) == 2, 'ngram size must equal to 2'
    if 'tfidf' in vectorizer.lower():
        char_vectorizer = TfidfVectorizer(
            sublinear_tf = True,
            strip_accents = 'unicode',
            analyzer = 'char',
            ngram_range = ngram,
        )
    elif 'count' in vectorizer.lower():
        char_vectorizer = CountVectorizer(
            strip_accents = 'unicode', analyzer = 'char', ngram_range = ngram
        )
    else:
        raise Exception('model not supported')
    output, keys = _generate_influencers()
    vectorizer = char_vectorizer.fit(output)
    vectorized = vectorizer.transform(output)
    return _FAST_SIMILARITY(vectorizer, vectorized, keys, is_influencers = True)



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import re
from .texts._tatabahasa import tatabahasa_dict, hujung, permulaan
from ._utils import _tag_class
from ._utils._paths import PATH_POS, S3_PATH_POS


def available_deep_model():
    """
    List available deep learning entities models, ['concat', 'bahdanau', 'luong', 'entity-network', 'attention']
    """
    return ['concat', 'bahdanau', 'luong', 'entity-network', 'attention']


def _naive_POS_word(word):
    for key, vals in tatabahasa_dict.items():
        if word in vals:
            return (key, word)
    try:
        if len(re.findall(r'^(.*?)(%s)$' % ('|'.join(hujung[:1])), i)[0]) > 1:
            return ('KJ', word)
    except:
        pass
    try:
        if (
            len(re.findall(r'^(.*?)(%s)' % ('|'.join(permulaan[:-4])), word)[0])
            > 1
        ):
            return ('KJ', word)
    except Exception as e:
        pass
    if len(word) > 2:
        return ('KN', word)
    else:
        return ('', word)


def naive(string):
    """
    Recognize POS in a string using Regex.

    Parameters
    ----------
    string: str

    Returns
    -------
    string : tokenized string with POS related
    """
    assert isinstance(string, str), 'input must be a string'
    string = string.lower()
    results = []
    for i in string.split():
        results.append(_naive_POS_word(i))
    return results


def crf(validate = True):
    """
    Load CRF POS Recognition model.

    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    CRF : malaya.sklearn_model.CRF class
    """
    return _tag_class.crf(
        PATH_POS, S3_PATH_POS, 'pos', is_lower = False, validate = validate
    )


def deep_model(model = 'concat', validate = True):
    """
    Load deep learning POS Recognition model.

    Parameters
    ----------
    model : str, optional (default='bahdanau')
        Model architecture supported. Allowed values:

        * ``'concat'`` - Concating character and word embedded for BiLSTM
        * ``'bahdanau'`` - Concating character and word embedded including Bahdanau Attention for BiLSTM
        * ``'luong'`` - Concating character and word embedded including Luong Attention for BiLSTM
        * ``'entity-network'`` - Concating character and word embedded on hybrid Entity-Network and RNN
        * ``'attention'`` - Concating character and word embedded with self-attention for BiLSTM
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    TAGGING: malaya.tensorflow_model.TAGGING class
    """
    return _tag_class.deep_model(
        PATH_POS,
        S3_PATH_POS,
        'pos',
        model = model,
        is_lower = False,
        validate = validate,
    )



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import tensorflow as tf
import json
import pickle
from ._utils import check_file, load_graph, check_available
from ..stem import _classification_textcleaning_stemmer
from .._models._sklearn_model import USER_XGB, USER_BAYES
from .._models._tensorflow_model import SOFTMAX, SPARSE_SOFTMAX


def sparse_deep_model(
    path,
    s3_path,
    class_name,
    label,
    output_size,
    embedded_size = 64,
    model = 'fast-text-char',
    validate = True,
):
    import os

    assert isinstance(model, str), 'model must be a string'
    model = model.lower()
    if model == 'fast-text-char':
        if validate:
            check_file(path[model], s3_path[model])
        else:
            if not check_available(path[model]):
                raise Exception(
                    '%s/%s is not available, please `validate = True`'
                    % (class_name, model)
                )
        try:
            with open(path[model]['vector'], 'rb') as fopen:
                vector = pickle.load(fopen)

            return SPARSE_SOFTMAX(
                os.path.dirname(path[model]['model']),
                vector,
                label,
                output_size,
                embedded_size,
                len(vector.vocabulary_),
            )
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('%s/%s') and try again"
                % (class_name, model)
            )
    else:
        raise Exception(
            'model subjectivity not supported, please check supported models from malaya.%s.available_sparse_deep_model()'
            % (class_name)
        )


def deep_model(
    path, s3_path, class_name, label, model = 'luong', validate = True
):
    assert isinstance(model, str), 'model must be a string'
    model = model.lower()
    if model == 'fast-text':
        if validate:
            check_file(path['fast-text'], s3_path['fast-text'])
        else:
            if not check_available(path['fast-text']):
                raise Exception(
                    '%s/%s is not available, please `validate = True`'
                    % (class_name, model)
                )
        try:
            with open(path['fast-text']['setting'], 'r') as fopen:
                dictionary = json.load(fopen)['dictionary']
            with open(path['fast-text']['pickle'], 'rb') as fopen:
                ngram = pickle.load(fopen)
            g = load_graph(path['fast-text']['model'])
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('%s/%s') and try again"
                % (class_name, model)
            )
        return SOFTMAX(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            ngram = ngram,
            label = label,
        )
    elif model == 'hierarchical':
        if validate:
            check_file(path['hierarchical'], s3_path['hierarchical'])
        else:
            if not check_available(path['hierarchical']):
                raise Exception(
                    '%s/%s is not available, please `validate = True`'
                    % (class_name, model)
                )
        try:
            with open(path['hierarchical']['setting'], 'r') as fopen:
                dictionary = json.load(fopen)['dictionary']
            g = load_graph(path['hierarchical']['model'])
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('%s/%s') and try again"
                % (class_name, model)
            )
        return SOFTMAX(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            alphas = g.get_tensor_by_name('import/alphas:0'),
            label = label,
        )
    elif model in ['bahdanau', 'luong']:
        if validate:
            check_file(path[model], s3_path[model])
        else:
            if not check_available(path[model]):
                raise Exception(
                    '%s/%s is not available, please `validate = True`'
                    % (class_name, model)
                )
        try:
            with open(path[model]['setting'], 'r') as fopen:
                dictionary = json.load(fopen)['dictionary']
            g = load_graph(path[model]['model'])
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('%s/%s') and try again"
                % (class_name, model)
            )
        return SOFTMAX(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            alphas = g.get_tensor_by_name('import/alphas:0'),
            label = label,
        )
    elif model == 'bidirectional':
        if validate:
            check_file(path['bidirectional'], s3_path['bidirectional'])
        else:
            if not check_available(path['bidirectional']):
                raise Exception(
                    '%s/%s is not available, please `validate = True`'
                    % (class_name, model)
                )
        try:
            with open(path['bidirectional']['setting'], 'r') as fopen:
                dictionary = json.load(fopen)
            g = load_graph(path['bidirectional']['model'])
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('%s/%s') and try again"
                % (class_name, model)
            )
        return SOFTMAX(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            label = label,
        )
    elif model == 'bert':
        if validate:
            check_file(path['bert'], s3_path['bert'])
        else:
            if not check_available(path['bert']):
                raise Exception(
                    '%s/%s is not available, please `validate = True`'
                    % (class_name, model)
                )
        try:
            with open(path['bert']['setting'], 'r') as fopen:
                dictionary = json.load(fopen)
            g = load_graph(path['bert']['model'])
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('%s/%s') and try again"
                % (class_name, model)
            )
        return SOFTMAX(
            g.get_tensor_by_name('import/Placeholder_input_ids:0'),
            g.get_tensor_by_name('import/loss/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            input_mask = g.get_tensor_by_name(
                'import/Placeholder_input_mask:0'
            ),
            segment_ids = g.get_tensor_by_name(
                'import/Placeholder_segment_ids:0'
            ),
            is_training = g.get_tensor_by_name(
                'import/Placeholder_is_training:0'
            ),
            label = label,
        )
    elif model == 'entity-network':
        if validate:
            check_file(path['entity-network'], s3_path['entity-network'])
        else:
            if not check_available(path['entity-network']):
                raise Exception(
                    '%s/%s is not available, please `validate = True`'
                    % (class_name, model)
                )
        try:
            with open(path['entity-network']['setting'], 'r') as fopen:
                dictionary = json.load(fopen)
            g = load_graph(path['entity-network']['model'])
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('%s/%s') and try again"
                % (class_name, model)
            )
        return SOFTMAX(
            g.get_tensor_by_name('import/Placeholder_question:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            dropout_keep_prob = g.get_tensor_by_name(
                'import/Placeholder_dropout_keep_prob:0'
            ),
            story = g.get_tensor_by_name('import/Placeholder_story:0'),
            label = label,
        )
    else:
        raise Exception(
            'model subjectivity not supported, please check supported models from malaya.%s.available_deep_model()'
            % (class_name)
        )


def multinomial(path, s3_path, class_name, label, validate = True):
    if validate:
        check_file(path['multinomial'], s3_path['multinomial'])
    else:
        if not check_available(path['multinomial']):
            raise Exception(
                '%s/multinomial is not available, please `validate = True`'
                % (class_name)
            )
    try:
        with open(path['multinomial']['model'], 'rb') as fopen:
            multinomial = pickle.load(fopen)
        with open(path['multinomial']['vector'], 'rb') as fopen:
            vectorize = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('%s/multinomial') and try again"
            % (class_name)
        )
    return USER_BAYES(
        multinomial,
        label,
        vectorize,
        cleaning = _classification_textcleaning_stemmer,
    )


def xgb(path, s3_path, class_name, label, validate = True):
    if validate:
        check_file(path['xgb'], s3_path['xgb'])
    else:
        if not check_available(path['xgb']):
            raise Exception(
                '%s/xgb is not available, please `validate = True`'
                % (class_name)
            )
    try:
        with open(path['xgb']['model'], 'rb') as fopen:
            xgb = pickle.load(fopen)
        with open(path['xgb']['vector'], 'rb') as fopen:
            vectorize = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('%s/xgb') and try again"
            % (class_name)
        )
    return USER_XGB(
        xgb, label, vectorize, cleaning = _classification_textcleaning_stemmer
    )



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from collections import defaultdict
from itertools import chain
from six import string_types


class DependencyGraph(object):
    def __init__(
        self,
        tree_str = None,
        cell_extractor = None,
        zero_based = False,
        cell_separator = None,
        top_relation_label = 'ROOT',
    ):
        self.nodes = defaultdict(
            lambda: {
                'address': None,
                'word': None,
                'lemma': None,
                'ctag': None,
                'tag': None,
                'feats': None,
                'head': None,
                'deps': defaultdict(list),
                'rel': None,
            }
        )

        self.nodes[0].update({'ctag': 'TOP', 'tag': 'TOP', 'address': 0})

        self.root = None

        if tree_str:
            self._parse(
                tree_str,
                cell_extractor = cell_extractor,
                zero_based = zero_based,
                cell_separator = cell_separator,
                top_relation_label = top_relation_label,
            )

    def remove_by_address(self, address):
        del self.nodes[address]

    def redirect_arcs(self, originals, redirect):
        for node in self.nodes.values():
            new_deps = []
            for dep in node['deps']:
                if dep in originals:
                    new_deps.append(redirect)
                else:
                    new_deps.append(dep)
            node['deps'] = new_deps

    def add_arc(self, head_address, mod_address):
        relation = self.nodes[mod_address]['rel']
        self.nodes[head_address]['deps'].setdefault(relation, [])
        self.nodes[head_address]['deps'][relation].append(mod_address)

    def connect_graph(self):
        for node1 in self.nodes.values():
            for node2 in self.nodes.values():
                if (
                    node1['address'] != node2['address']
                    and node2['rel'] != 'TOP'
                ):
                    relation = node2['rel']
                    node1['deps'].setdefault(relation, [])
                    node1['deps'][relation].append(node2['address'])

    def get_by_address(self, node_address):
        """Return the node with the given address."""
        return self.nodes[node_address]

    def contains_address(self, node_address):
        """
        Returns true if the graph contains a node with the given node
        address, false otherwise.
        """
        return node_address in self.nodes

    def to_dot(self):
        s = 'digraph G{\n'
        s += 'edge [dir=forward]\n'
        s += 'node [shape=plaintext]\n'

        for node in sorted(self.nodes.values(), key = lambda v: v['address']):
            s += '\n%s [label="%s (%s)"]' % (
                node['address'],
                node['address'],
                node['word'],
            )
            for rel, deps in node['deps'].items():
                for dep in deps:
                    if rel is not None:
                        s += '\n%s -> %s [label="%s"]' % (
                            node['address'],
                            dep,
                            rel,
                        )
                    else:
                        s += '\n%s -> %s ' % (node['address'], dep)
        s += '\n}'

        return s

    def left_children(self, node_index):
        children = chain.from_iterable(self.nodes[node_index]['deps'].values())
        index = self.nodes[node_index]['address']
        return sum(1 for c in children if c < index)

    def right_children(self, node_index):
        children = chain.from_iterable(self.nodes[node_index]['deps'].values())
        index = self.nodes[node_index]['address']
        return sum(1 for c in children if c > index)

    def add_node(self, node):
        if not self.contains_address(node['address']):
            self.nodes[node['address']].update(node)

    def _parse(
        self,
        input_,
        cell_extractor = None,
        zero_based = False,
        cell_separator = None,
        top_relation_label = 'ROOT',
    ):
        def extract_3_cells(cells, index):
            word, tag, head = cells
            return index, word, word, tag, tag, '', head, ''

        def extract_4_cells(cells, index):
            word, tag, head, rel = cells
            return index, word, word, tag, tag, '', head, rel

        def extract_7_cells(cells, index):
            line_index, word, lemma, tag, _, head, rel = cells
            try:
                index = int(line_index)
            except ValueError:
                # index can't be parsed as an integer, use default
                pass
            return index, word, lemma, tag, tag, '', head, rel

        def extract_10_cells(cells, index):
            line_index, word, lemma, ctag, tag, feats, head, rel, _, _ = cells
            try:
                index = int(line_index)
            except ValueError:
                # index can't be parsed as an integer, use default
                pass
            return index, word, lemma, ctag, tag, feats, head, rel

        extractors = {
            3: extract_3_cells,
            4: extract_4_cells,
            7: extract_7_cells,
            10: extract_10_cells,
        }

        if isinstance(input_, string_types):
            input_ = (line for line in input_.split('\n'))

        lines = (l.rstrip() for l in input_)
        lines = (l for l in lines if l)

        cell_number = None
        for index, line in enumerate(lines, start = 1):
            cells = line.split(cell_separator)
            if cell_number is None:
                cell_number = len(cells)
            else:
                assert cell_number == len(cells)

            if cell_extractor is None:
                try:
                    cell_extractor = extractors[cell_number]
                except KeyError:
                    raise ValueError(
                        'Number of tab-delimited fields ({0}) not supported by '
                        'CoNLL(10) or Malt-Tab(4) format'.format(cell_number)
                    )

            try:
                index, word, lemma, ctag, tag, feats, head, rel = cell_extractor(
                    cells, index
                )
            except (TypeError, ValueError):
                word, lemma, ctag, tag, feats, head, rel = cell_extractor(cells)

            if head == '_':
                continue

            head = int(head)
            if zero_based:
                head += 1

            self.nodes[index].update(
                {
                    'address': index,
                    'word': word,
                    'lemma': lemma,
                    'ctag': ctag,
                    'tag': tag,
                    'feats': feats,
                    'head': head,
                    'rel': rel,
                }
            )

            # Make sure that the fake root node has labeled dependencies.
            if (cell_number == 3) and (head == 0):
                rel = top_relation_label
            self.nodes[head]['deps'][rel].append(index)

        if self.nodes[0]['deps'][top_relation_label]:
            root_address = self.nodes[0]['deps'][top_relation_label][0]
            self.root = self.nodes[root_address]
            self.top_relation_label = top_relation_label
        else:
            warnings.warn(
                "The graph doesn't contain a node "
                'that depends on the root element.'
            )

    def _word(self, node, filter = True):
        w = node['word']
        if filter:
            if w != ',':
                return w
        return w

    def triples(self, node = None):
        """
        Extract dependency triples of the form:
        ((head word, head tag), rel, (dep word, dep tag))
        """

        if not node:
            node = self.root

        head = (node['word'], node['ctag'])
        for i in sorted(chain.from_iterable(node['deps'].values())):
            dep = self.get_by_address(i)
            yield (head, dep['rel'], (dep['word'], dep['ctag']))
            for triple in self.triples(node = dep):
                yield triple

    def _hd(self, i):
        try:
            return self.nodes[i]['head']
        except IndexError:
            return None

    def _rel(self, i):
        try:
            return self.nodes[i]['rel']
        except IndexError:
            return None

    def contains_cycle(self):
        """Check whether there are cycles.

        >>> dg = DependencyGraph(treebank_data)
        >>> dg.contains_cycle()
        False

        >>> cyclic_dg = DependencyGraph()
        >>> top = {'word': None, 'deps': [1], 'rel': 'TOP', 'address': 0}
        >>> child1 = {'word': None, 'deps': [2], 'rel': 'NTOP', 'address': 1}
        >>> child2 = {'word': None, 'deps': [4], 'rel': 'NTOP', 'address': 2}
        >>> child3 = {'word': None, 'deps': [1], 'rel': 'NTOP', 'address': 3}
        >>> child4 = {'word': None, 'deps': [3], 'rel': 'NTOP', 'address': 4}
        >>> cyclic_dg.nodes = {
        ...     0: top,
        ...     1: child1,
        ...     2: child2,
        ...     3: child3,
        ...     4: child4,
        ... }
        >>> cyclic_dg.root = top

        >>> cyclic_dg.contains_cycle()
        [3, 1, 2, 4]

        """
        distances = {}

        for node in self.nodes.values():
            for dep in node['deps']:
                key = tuple([node['address'], dep])
                distances[key] = 1

        for _ in self.nodes:
            new_entries = {}

            for pair1 in distances:
                for pair2 in distances:
                    if pair1[1] == pair2[0]:
                        key = tuple([pair1[0], pair2[1]])
                        new_entries[key] = distances[pair1] + distances[pair2]

            for pair in new_entries:
                distances[pair] = new_entries[pair]
                if pair[0] == pair[1]:
                    path = self.get_cycle_path(
                        self.get_by_address(pair[0]), pair[0]
                    )
                    return path

        return False

    def get_cycle_path(self, curr_node, goal_node_index):
        for dep in curr_node['deps']:
            if dep == goal_node_index:
                return [curr_node['address']]
        for dep in curr_node['deps']:
            path = self.get_cycle_path(
                self.get_by_address(dep), goal_node_index
            )
            if len(path) > 0:
                path.insert(0, curr_node['address'])
                return path
        return []

    def to_conll(self, style):

        if style == 3:
            template = '{word}\t{tag}\t{head}\n'
        elif style == 4:
            template = '{word}\t{tag}\t{head}\t{rel}\n'
        elif style == 10:
            template = '{i}\t{word}\t{lemma}\t{ctag}\t{tag}\t{feats}\t{head}\t{rel}\t_\t_\n'
        else:
            raise ValueError(
                'Number of tab-delimited fields ({0}) not supported by '
                'CoNLL(10) or Malt-Tab(4) format'.format(style)
            )

        return ''.join(
            template.format(i = i, **node)
            for i, node in sorted(self.nodes.items())
            if node['tag'] != 'TOP'
        )

    def to_graphvis(self):
        try:
            from graphviz import Source
        except:
            raise Exception(
                'graphiz not installed. Please install it and try again.'
            )
        return Source(self.to_dot())

    def to_networkx(self):
        try:
            import networkx
        except:
            raise Exception(
                'networkx not installed. Please install it and try again.'
            )

        nx_nodelist = list(range(1, len(self.nodes)))
        nx_edgelist = [
            (n, self._hd(n), self._rel(n)) for n in nx_nodelist if self._hd(n)
        ]
        self.nx_labels = {}
        for n in nx_nodelist:
            self.nx_labels[n] = self.nodes[n]['word']

        g = networkx.MultiDiGraph()
        g.add_nodes_from(nx_nodelist)
        g.add_edges_from(nx_edgelist)

        return g



import numpy as np
import itertools

_accepted_pos = [
    'ADJ',
    'ADP',
    'ADV',
    'ADX',
    'CCONJ',
    'DET',
    'NOUN',
    'NUM',
    'PART',
    'PRON',
    'PROPN',
    'SCONJ',
    'SYM',
    'VERB',
    'X',
]
_accepted_entities = [
    'OTHER',
    'law',
    'location',
    'organization',
    'person',
    'quantity',
    'time',
    'event',
]


def cluster_words(list_words):
    """
    cluster similar words based on structure, eg, ['mahathir mohamad', 'mahathir'] = ['mahathir mohamad']

    Parameters
    ----------
    list_words : list of str

    Returns
    -------
    string: list of clustered words
    """
    assert isinstance(list_words, list) and isinstance(
        list_words[0], str
    ), 'input must be list of strings'
    dict_words = {}
    for word in list_words:
        found = False
        for key in dict_words.keys():
            if word in key or any(
                [word in inside for inside in dict_words[key]]
            ):
                dict_words[key].append(word)
                found = True
            if key in word:
                dict_words[key].append(word)
        if not found:
            dict_words[word] = [word]
    results = []
    for key, words in dict_words.items():
        results.append(max(list(set([key] + words)), key = len))
    return list(set(results))


def _pad_sequence(
    sequence,
    n,
    pad_left = False,
    pad_right = False,
    left_pad_symbol = None,
    right_pad_symbol = None,
):
    sequence = iter(sequence)
    if pad_left:
        sequence = itertools.chain((left_pad_symbol,) * (n - 1), sequence)
    if pad_right:
        sequence = itertools.chain(sequence, (right_pad_symbol,) * (n - 1))
    return sequence


def ngrams(
    sequence,
    n,
    pad_left = False,
    pad_right = False,
    left_pad_symbol = None,
    right_pad_symbol = None,
):
    """
    generate ngrams

    Parameters
    ----------
    sequence : list of str
        list of tokenize words
    n : int
        ngram size

    Returns
    -------
    ngram: list
    """
    sequence = _pad_sequence(
        sequence, n, pad_left, pad_right, left_pad_symbol, right_pad_symbol
    )

    history = []
    while n > 1:
        try:
            next_item = next(sequence)
        except StopIteration:
            return
        history.append(next_item)
        n -= 1
    for item in sequence:
        history.append(item)
        yield tuple(history)
        del history[0]


def pos_entities_ngram(
    result_pos,
    result_entities,
    ngram = (1, 3),
    accept_pos = ['NOUN', 'PROPN', 'VERB'],
    accept_entities = ['law', 'location', 'organization', 'person', 'time'],
):
    """
    generate ngrams

    Parameters
    ----------
    result_pos : list of tuple
        result from POS recognition
    result_entities : list of tuple
        result of Entities recognition
    ngram : tuple
        ngram sizes
    accept_pos : list of str
        accepted POS elements
    accept_entities : list of str
        accept entities elements

    Returns
    -------
    result: list
    """
    assert isinstance(result_pos, list) and isinstance(
        result_pos[0], tuple
    ), 'result_pos must be list of tuple'
    assert isinstance(result_entities, list) and isinstance(
        result_entities[0], tuple
    ), 'result_entities must be list of tuple'
    assert isinstance(ngram, tuple), 'ngram must be a tuple'
    assert len(ngram) == 2, 'ngram size must equal to 2'
    assert isinstance(accept_pos, list), 'accept_pos must be a list'
    assert isinstance(accept_entities, list), 'accept_entites must be a list'
    assert all(
        [i in _accepted_pos for i in accept_pos]
    ), 'accept_pos must be a subset or equal of supported POS, please run malaya.describe_pos() to get supported POS'
    assert all(
        [i in _accepted_entities for i in accept_entities]
    ), 'accept_entites must be a subset or equal of supported entities, please run malaya.describe_entities() to get supported entities'

    words = []
    sentences = []
    for no in range(len(result_pos)):
        if (
            result_pos[no][1] in accept_pos
            or result_entities[no][1] in accept_entities
        ):
            words.append(result_pos[no][0])
    for gram in range(ngram[0], ngram[1] + 1, 1):
        gram_words = list(ngrams(words, gram))
        for sentence in gram_words:
            sentences.append(' '.join(sentence))
    return list(set(sentences))


def sentence_ngram(sentence, ngram = (1, 3)):
    """
    generate ngram for a text

    Parameters
    ----------
    sentence: str
    ngram : tuple
        ngram sizes

    Returns
    -------
    result: list
    """
    assert (sentence, str), 'sentence must be a string'
    assert isinstance(ngram, tuple), 'ngram must be a tuple'
    assert len(ngram) == 2, 'ngram size must equal to 2'
    words = sentence.split()
    sentences = []
    for gram in range(ngram[0], ngram[1] + 1, 1):
        gram_words = list(ngrams(words, gram))
        for sentence in gram_words:
            sentences.append(' '.join(sentence))
    return list(set(sentences))


def cluster_pos(result):
    """
    cluster similar POS

    Parameters
    ----------
    result: list

    Returns
    -------
    result: list
    """
    output = {
        'ADJ': [],
        'ADP': [],
        'ADV': [],
        'ADX': [],
        'CCONJ': [],
        'DET': [],
        'NOUN': [],
        'NUM': [],
        'PART': [],
        'PRON': [],
        'PROPN': [],
        'SCONJ': [],
        'SYM': [],
        'VERB': [],
        'X': [],
    }
    last_label, words = None, []
    for word, label in result:
        if last_label != label and last_label:
            joined = ' '.join(words)
            if joined not in output[last_label]:
                output[last_label].append(joined)
            words = []
            last_label = label
            words.append(word)

        else:
            if not last_label:
                last_label = label
            words.append(word)
    return output


def cluster_entities(result):
    """
    cluster similar Entities

    Parameters
    ----------
    result: list

    Returns
    -------
    result: list
    """
    output = {
        'OTHER': [],
        'law': [],
        'location': [],
        'organization': [],
        'person': [],
        'quantity': [],
        'time': [],
        'event': [],
    }
    last_label, words = None, []
    for word, label in result:
        if last_label != label and last_label:
            joined = ' '.join(words)
            if joined not in output[last_label]:
                output[last_label].append(joined)
            words = []
            last_label = label
            words.append(word)

        else:
            if not last_label:
                last_label = label
            words.append(word)
    return output



import malaya

def test_topics_influencers_short():
    news = 'najib razak dan mahathir mengalami masalah air di kemamam terengganu'
    assert len(malaya.get_topics(news)) > 0
    assert len(malaya.get_influencers(news)) > 0

def test_topics_influencers_long():
    news = 'ikat penyedia perkhidmatan jalur lebar Telekom Malaysia (TM) perlu mencari jalan penyelesaian bagi meningkatkan akses capaian Internet ke seluruh negara, kata Menteri Komunikasi dan Multimedia, Gobind Singh Deo. Beliau berkata menjadi dasar kerajaan untuk membekalkan akses Internet jalur lebar kepada semua dan memberi penekanan kepada kualiti perkhidmatan yang terbaik. "Dasar kerajaan untuk bekalkan akses kepada semua bukan sekadar pembekalan sahaja tetapi beri penekanan kepada kualiti perkhidmatan yang baik dan dapat bersaing dengan negara lain pada tahap antarabangsa," kata Gobind Singh menerusi catatan di laman rasmi Twitter beliau, malam tadi. Beliau berkata demikian sebagai respons terhadap aduan beberapa pengguna Twitter berhubung akses Internet yang masih tidak stabil serta harga yang tidak berpatutan di beberapa lokasi di seluruh negara.'
    assert len(malaya.get_topics(news)) > 0
    assert len(malaya.get_influencers(news)) > 0



GO = 0
PAD = 1
EOS = 2
UNK = 3

tanya_list = [
    'kenapa',
    'bila',
    'siapa',
    'mengapa',
    'apa',
    'bagaimana',
    'berapa',
    'mana',
]
perintah_list = ['jangan', 'sila', 'tolong', 'harap', 'usah', 'jemput', 'minta']
pangkal_list = [
    'maka',
    'alkisah',
    'arakian',
    'syahdah',
    'adapun',
    'bermula',
    'kalakian',
]
bantu_list = [
    'akan',
    'telah',
    'boleh',
    'mesti',
    'belum',
    'sudah',
    'dapat',
    'masih',
    'harus',
    'hendak',
]
penguat_list = [
    'paling',
    'agak',
    'sungguh',
    'amat',
    'terlalu',
    'nian',
    'benar',
    'paling',
]
penegas_list = ['jua', 'juga', 'sahaja', 'hanya', 'memang', 'lagi', 'pun']
nafi_list = ['bukan', 'tidak', 'tak', 'tiada', 'tidaklah', 'tidakkah']
pemeri_list = ['ialah', 'adalah']
sendi_list = [
    'akan',
    'kepada',
    'terhadap',
    'bagi',
    'untuk',
    'dari',
    'daripada',
    'di',
    'dengan',
    'hingga',
    'sampai',
    'ke',
    'kepada',
    'oleh',
    'pada',
    'sejak',
    'seperti',
    'umpama',
    'bak',
    'tentang',
    'laksanabagai',
    'semenjak',
    'dalam',
    'antara',
]
pembenar_list = ['ya', 'benar', 'betul']
nombor_list = [
    'satu',
    'dua',
    'tiga',
    'empat',
    'lima',
    'enam',
    'tujuh',
    'lapan',
    'sembilan',
    'kosong',
]
suku_bilangan_list = ['per', 'suku', 'setengah', 'separuh', 'tiga suku']
pisahan_list = ['setiap', 'tiap']
keterangan_list = [
    'begitu',
    'begini',
    'demikian',
    'perlahan',
    'cepat',
    'lena',
    'akan',
    'sedang',
    'belum',
    'telah',
    'sekarang',
    'sebentar',
    'semalam',
    'mungkin',
    'agak',
    'barangkali',
    'pasti',
    'tentu',
    'sudah',
    'selalu',
    'kadang',
    'acapkali',
    'sesekali',
    'yang',
]
arah_list = [
    'atas',
    'bawah',
    'tepi',
    'antara',
    'hadapan',
    'utara',
    'sisi',
    'luar',
]
hubung_list = [
    'agar',
    'apabila',
    'atau',
    'bahawa',
    'dan',
    'hingga',
    'jika',
    'jikalau',
    'kecuali',
    'kerana',
    'lalu',
    'manakala',
    'sambil',
    'serta',
    'semenjak',
    'sementara',
    'sungguhpun',
    'supaya',
    'walaupun',
    'tetapi',
    'berkenan',
    'berkenaan',
]
gantinama_list = [
    'aku',
    'saya',
    'hamba',
    'patik',
    'beta',
    'kami',
    'kita',
    'anda',
    'awak',
    'engkau',
    'tuanku',
    'kalian',
    'kamu',
    'baginda',
    'beliau',
    'mereka',
    'ini',
    'itu',
    'sini',
    'situ',
    'sana',
    'kini',
    'dia',
    'rm',
]

permulaan = [
    'bel',
    'se',
    'ter',
    'men',
    'meng',
    'mem',
    'memper',
    'di',
    'pe',
    'me',
    'ke',
    'ber',
    'pen',
    'per',
]

hujung = ['kan', 'kah', 'lah', 'tah', 'nya', 'an', 'wan', 'wati', 'ita']
alphabet = 'qwertyuiopasdfghjklzxcvbnm'
consonants = 'bcdfghjklmnpqrstvwxyz'
vowels = 'aeiou'

sounds = {
    'x': 'tidak',
    'y': 'kenapa',
    'n': 'dan',
    'g': 'pergi',
    's': 'seperti',
    'd': 'di',
    'k': 'ok',
    'u': 'awak',
    't': 'nanti',
    'p': 'pergi',
    'wai': 'kenapa',
}

tatabahasa_dict = {
    'KT': tanya_list,
    'KP': perintah_list,
    'KPA': pangkal_list,
    'KB': bantu_list,
    'KPENGUAT': penguat_list,
    'KPENEGAS': penegas_list,
    'NAFI': nafi_list,
    'KPEMERI': pemeri_list,
    'KS': sendi_list,
    'KPEMBENAR': pembenar_list,
    'NO': nombor_list,
    'SUKU': suku_bilangan_list,
    'PISAHAN': pisahan_list,
    'KETERANGAN': keterangan_list,
    'ARAH': arah_list,
    'KH': hubung_list,
    'GN': gantinama_list,
}

stopword_tatabahasa = list(
    set(
        tanya_list
        + perintah_list
        + pangkal_list
        + bantu_list
        + penguat_list
        + penegas_list
        + nafi_list
        + pemeri_list
        + sendi_list
        + pembenar_list
        + nombor_list
        + suku_bilangan_list
        + pisahan_list
        + keterangan_list
        + arah_list
        + hubung_list
        + gantinama_list
    )
)

rules_normalizer = {
    'experience': 'pengalaman',
    'bagasi': 'bagasi',
    'kg': 'kampung',
    'kilo': 'kilogram',
    'g': 'gram',
    'grm': 'gram',
    'k': 'okay',
    'abgkat': 'abang dekat',
    'abis': 'habis',
    'ade': 'ada',
    'adoi': 'aduh',
    'adoii': 'aduhh',
    'aerodarat': 'kapal darat',
    'agkt': 'angkat',
    'ahh': 'ah',
    'ailior': 'air liur',
    'airasia': 'air asia x',
    'airasiax': 'penerbangan',
    'airline': 'penerbangan',
    'airlines': 'penerbangan',
    'airport': 'lapangan terbang',
    'airpot': 'lapangan terbang',
    'aje': 'sahaja',
    'ajelah': 'sahajalah',
    'ajer': 'sahaja',
    'ak': 'aku',
    'aq': 'aku',
    'all': 'semua',
    'ambik': 'ambil',
    'amek': 'ambil',
    'amer': 'amir',
    'amik': 'ambil',
    'ana': 'saya',
    'angkt': 'angkat',
    'anual': 'tahunan',
    'apapun': 'apa pun',
    'ape': 'apa',
    'arab': 'arab',
    'area': 'kawasan',
    'aritu': 'hari itu',
    'ask': 'tanya',
    'astro': 'astro',
    'at': 'pada',
    'attitude': 'sikap',
    'babi': 'khinzir',
    'back': 'belakang',
    'bag': 'beg',
    'bang': 'abang',
    'bangla': 'bangladesh',
    'banyk': 'banyak',
    'bard': 'pujangga',
    'bargasi': 'bagasi',
    'bawak': 'bawa',
    'bawanges': 'bawang',
    'be': 'jadi',
    'behave': 'berkelakuan baik',
    'belagak': 'berlagak',
    'berdisiplin': 'berdisplin',
    'berenti': 'berhenti',
    'beskal': 'basikal',
    'bff': 'rakan karib',
    'bg': 'bagi',
    'bgi': 'bagi',
    'biase': 'biasa',
    'big': 'besar',
    'bike': 'basikal',
    'bile': 'bila',
    'binawe': 'binatang',
    'bini': 'isteri',
    'bkn': 'bukan',
    'bla': 'bila',
    'blom': 'belum',
    'bnyak': 'banyak',
    'body': 'tubuh',
    'bole': 'boleh',
    'boss': 'bos',
    'bowling': 'boling',
    'bpe': 'berapa',
    'brand': 'jenama',
    'brg': 'barang',
    'briefing': 'taklimat',
    'brng': 'barang',
    'bro': 'abang',
    'bru': 'baru',
    'bruntung': 'beruntung',
    'bsikal': 'basikal',
    'btnggjwb': 'bertanggungjawab',
    'btul': 'betul',
    'buatlh': 'buatlah',
    'buh': 'letak',
    'buka': 'buka',
    'but': 'tetapi',
    'bwk': 'bawa',
    'by': 'dengan',
    'byr': 'bayar',
    'bz': 'sibuk',
    'camera': 'kamera',
    'camni': 'macam ini',
    'cane': 'macam mana',
    'cant': 'tak boleh',
    'carakerja': 'cara kerja',
    'care': 'jaga',
    'cargo': 'kargo',
    'cctv': 'kamera litar tertutup',
    'celako': 'celaka',
    'cer': 'cerita',
    'cheap': 'murah',
    'check': 'semak',
    'ciput': 'sedikit',
    'cite': 'cerita',
    'citer': 'cerita',
    'ckit': 'sikit',
    'ckp': 'cakap',
    'class': 'kelas',
    'cm': 'macam',
    'cmni': 'macam ini',
    'cmpak': 'campak',
    'committed': 'komited',
    'company': 'syarikat',
    'complain': 'aduan',
    'corn': 'jagung',
    'couldnt': 'tak boleh',
    'cr': 'cari',
    'crew': 'krew',
    'cube': 'cuba',
    'cuma': 'cuma',
    'curinyaa': 'curinya',
    'cust': 'pelanggan',
    'customer': 'pelanggan',
    'd': 'di',
    'da': 'dah',
    'dn': 'dan',
    'dahh': 'dah',
    'damaged': 'rosak',
    'dapek': 'dapat',
    'day': 'hari',
    'dazrin': 'dazrin',
    'dbalingnya': 'dibalingnya',
    'de': 'ada',
    'deep': 'dalam',
    'deliberately': 'sengaja',
    'depa': 'mereka',
    'dessa': 'desa',
    'dgn': 'dengan',
    'dh': 'dah',
    'didunia': 'di dunia',
    'diorang': 'mereka',
    'diorng': 'mereka',
    'direct': 'secara terus',
    'diving': 'junam',
    'dkt': 'dekat',
    'dlempar': 'dilempar',
    'dlm': 'dalam',
    'dlt': 'padam',
    'dlu': 'dulu',
    'done': 'siap',
    'dont': 'jangan',
    'dorg': 'mereka',
    'dpermudhkn': 'dipermudahkan',
    'dpt': 'dapat',
    'dr': 'dari',
    'dri': 'dari',
    'dsb': 'dan sebagainya',
    'dy': 'dia',
    'educate': 'mendidik',
    'ensure': 'memastikan',
    'everything': 'semua',
    'ewahh': 'wah',
    'expect': 'sangka',
    'fb': 'facebook',
    'fired': 'pecat',
    'first': 'pertama',
    'fkr': 'fikir',
    'flight': 'kapal terbang',
    'for': 'untuk',
    'free': 'percuma',
    'friend': 'kawan',
    'fyi': 'untuk pengetahuan anda',
    'gantila': 'gantilah',
    'gantirugi': 'ganti rugi',
    'gentlemen': 'lelaki budiman',
    'gerenti': 'jaminan',
    'gile': 'gila',
    'gk': 'juga',
    'gnti': 'ganti',
    'go': 'pergi',
    'gomen': 'kerajaan',
    'goment': 'kerajaan',
    'good': 'baik',
    'ground': 'tanah',
    'guarno': 'macam mana',
    'hampa': 'mereka',
    'hampeh': 'teruk',
    'hanat': 'jahanam',
    'handle': 'kawal',
    'handling': 'kawalan',
    'hanta': 'hantar',
    'haritu': 'hari itu',
    'hate': 'benci',
    'have': 'ada',
    'hawau': 'celaka',
    'henpon': 'telefon',
    'heran': 'hairan',
    'him': 'dia',
    'his': 'dia',
    'hmpa': 'mereka',
    'hntr': 'hantar',
    'hotak': 'otak',
    'hr': 'hari',
    'i': 'saya',
    'hrga': 'harga',
    'hrp': 'harap',
    'hu': 'sedih',
    'humble': 'merendah diri',
    'ibon': 'ikon',
    'ichi': 'inci',
    'idung': 'hidung',
    'if': 'jika',
    'ig': 'instagram',
    'iklas': 'ikhlas',
    'improve': 'menambah baik',
    'in': 'masuk',
    'isn t': 'tidak',
    'isyaallah': 'insyallah',
    'ja': 'sahaja',
    'japan': 'jepun',
    'jd': 'jadi',
    'je': 'saja',
    'jee': 'saja',
    'jek': 'saja',
    'jepun': 'jepun',
    'jer': 'saja',
    'jerr': 'saja',
    'jez': 'saja',
    'jg': 'juga',
    'jgk': 'juga',
    'jgn': 'jangan',
    'jgnla': 'janganlah',
    'jibake': 'celaka',
    'jjur': 'jujur',
    'job': 'kerja',
    'jobscope': 'skop kerja',
    'jogja': 'jogjakarta',
    'jpam': 'jpam',
    'jth': 'jatuh',
    'jugak': 'juga',
    'ka': 'ke',
    'kalo': 'kalau',
    'kalu': 'kalau',
    'kang': 'nanti',
    'kantoi': 'temberang',
    'kasi': 'beri',
    'kat': 'dekat',
    'kbye': 'ok bye',
    'kearah': 'ke arah',
    'kecik': 'kecil',
    'keja': 'kerja',
    'keje': 'kerja',
    'kejo': 'kerja',
    'keksongan': 'kekosongan',
    'kemana': 'ke mana',
    'kene': 'kena',
    'kenekan': 'kenakan',
    'kesah': 'kisah',
    'ketempat': 'ke tempat',
    'kije': 'kerja',
    'kijo': 'kerja',
    'kiss': 'cium',
    'kite': 'kita',
    'kito': 'kita',
    'kje': 'kerja',
    'kjr': 'kerja',
    'kk': 'okay',
    'kmi': 'kami',
    'kt': 'kat',
    'tlg': 'tolong',
    'kl': 'kuala lumpur',
    'klai': 'kalau',
    'klau': 'kalau',
    'klia': 'klia',
    'klo': 'kalau',
    'klu': 'kalau',
    'kn': 'kan',
    'knapa': 'kenapa',
    'kne': 'kena',
    'ko': 'kau',
    'kompom': 'sah',
    'korang': 'kamu semua',
    'korea': 'korea',
    'korg': 'kamu semua',
    'kot': 'mungkin',
    'krja': 'kerja',
    'ksalahan': 'kesalahan',
    'kta': 'kita',
    'kuar': 'keluar',
    'kut': 'mungkin',
    'la': 'lah',
    'laa': 'lah',
    'lahabau': 'celaka',
    'lahanat': 'celaka',
    'lainda': 'lain dah',
    'lak': 'pula',
    'last': 'akhir',
    'le': 'lah',
    'leader': 'ketua',
    'leave': 'pergi',
    'ler': 'lah',
    'less': 'kurang',
    'letter': 'surat',
    'lg': 'lagi',
    'lgi': 'lagi',
    'lngsong': 'langsung',
    'lol': 'hehe',
    'lorr': 'lah',
    'low': 'rendah',
    'lps': 'lepas',
    'luggage': 'bagasi',
    'lumbe': 'lumba',
    'lyak': 'layak',
    'maap': 'maaf',
    'maapkan': 'maafkan',
    'mahai': 'mahal',
    'mampos': 'mampus',
    'mart': 'kedai',
    'mau': 'mahu',
    'mcm': 'macam',
    'mcmtu': 'macam itu',
    'memerlukn': 'memerlukan',
    'mengembirakan': 'menggembirakan',
    'mengmbilnyer': 'mengambilnya',
    'mengtasi': 'mengatasi',
    'mg': 'memang',
    'mihak': 'memihak',
    'min': 'admin',
    'mingu': 'minggu',
    'mintak': 'minta',
    'mjtuhkn': 'menjatuhkan',
    'mkyong': 'mak yong',
    'mlibatkn': 'melibatkan',
    'mmg': 'memang',
    'mmnjang': 'memanjang',
    'mmpos': 'mampus',
    'mn': 'mana',
    'mna': 'mana',
    'mntak': 'minta',
    'mntk': 'minta',
    'mnyusun': 'menyusun',
    'mood': 'suasana',
    'most': 'paling',
    'mr': 'tuan',
    'msa': 'masa',
    'msia': 'malaysia',
    'mst': 'mesti',
    'mu': 'awak',
    'much': 'banyak',
    'muko': 'muka',
    'mum': 'emak',
    'n': 'dan',
    'nah': 'nah',
    'nanny': 'nenek',
    'napo': 'kenapa',
    'nati': 'nanti',
    'ngan': 'dengan',
    'ngn': 'dengan',
    'ni': 'ini',
    'nie': 'ini',
    'nii': 'ini',
    'nk': 'nak',
    'nmpk': 'nampak',
    'nye': 'nya',
    'ofis': 'pejabat',
    'ohh': 'oh',
    'oii': 'hoi',
    'one': 'satu',
    'online': 'dalam talian',
    'or': 'atau',
    'org': 'orang',
    'orng': 'orang',
    'otek': 'otak',
    'p': 'pergi',
    'paid': 'dah bayar',
    'palabana': 'kepala otak',
    'pasni': 'lepas ini',
    'passengers': 'penumpang',
    'passengger': 'penumpang',
    'pastu': 'lepas itu',
    'pd': 'pada',
    'pegi': 'pergi',
    'pekerje': 'pekerja',
    'pekrja': 'pekerja',
    'perabih': 'perabis',
    'perkerja': 'pekerja',
    'pg': 'pergi',
    'phuii': 'puih',
    'pikir': 'fikir',
    'pilot': 'juruterbang',
    'pk': 'fikir',
    'pkerja': 'pekerja',
    'pkerjaan': 'pekerjaan',
    'pki': 'pakai',
    'please': 'tolong',
    'pls': 'tolong',
    'pn': 'pun',
    'pnh': 'pernah',
    'pnt': 'penat',
    'pnya': 'punya',
    'pon': 'pun',
    'priority': 'keutamaan',
    'properties': 'harta benda',
    'ptugas': 'petugas',
    'pub': 'kelab malam',
    'pulak': 'pula',
    'puye': 'punya',
    'pwrcuma': 'percuma',
    'pyahnya': 'payahnya',
    'quality': 'kualiti',
    'quit': 'keluar',
    'ramly': 'ramly',
    'rege': 'harga',
    'reger': 'harga',
    'report': 'laporan',
    'resigned': 'meletakkan jawatan',
    'respect': 'hormat',
    'rizal': 'rizal',
    'rosak': 'rosak',
    'rosok': 'rosak',
    'rse': 'rasa',
    'sacked': 'buang',
    'sado': 'tegap',
    'salute': 'sanjung',
    'sam': 'sama',
    'same': 'sama',
    'samp': 'sampah',
    'sbb': 'sebab',
    'sbgai': 'sebagai',
    'sblm': 'sebelum',
    'sblum': 'sebelum',
    'sbnarnya': 'sebenarnya',
    'sbum': 'sebelum',
    'sdg': 'sedang',
    'sebb': 'sebab',
    'sebijik': 'sebiji',
    'see': 'lihat',
    'seen': 'dilihat',
    'selangor': 'selangor',
    'selfie': 'swafoto',
    'sempoi': 'cantik',
    'senaraihitam': 'senarai hitam',
    'seorg': 'seorang',
    'service': 'perkhidmatan',
    'sgt': 'sangat',
    'shared': 'kongsi',
    'shirt': 'kemeja',
    'shut': 'tutup',
    'sib': 'nasib',
    'skali': 'sekali',
    'sket': 'sikit',
    'sma': 'sama',
    'smoga': 'semoga',
    'smpoi': 'cantik',
    'sndiri': 'sendiri',
    'sndr': 'sendiri',
    'sndri': 'sendiri',
    'sne': 'sana',
    'so': 'jadi',
    'sop': 'tatacara pengendalian piawai',
    'sorang': 'seorang',
    'spoting': 'pembintikan',
    'sronok': 'seronok',
    'ssh': 'susah',
    'staff': 'staf',
    'standing': 'berdiri',
    'start': 'mula',
    'steady': 'mantap',
    'stiap': 'setiap',
    'stress': 'stres',
    'student': 'pelajar',
    'study': 'belajar',
    'studycase': 'kajian kes',
    'sure': 'pasti',
    'sykt': 'syarikat',
    'tah': 'entah',
    'taik': 'tahi',
    'takan': 'tak akan',
    'takat': 'setakat',
    'takde': 'tak ada',
    'takkan': 'tak akan',
    'taknak': 'tak nak',
    'tang': 'tentang',
    'tanggungjawab': 'bertanggungjawab',
    'taraa': 'sementara',
    'tau': 'tahu',
    'tbabit': 'terbabit',
    'team': 'pasukan',
    'terbaekk': 'terbaik',
    'teruknye': 'teruknya',
    'tgk': 'tengok',
    'that': 'itu',
    'thinking': 'fikir',
    'those': 'itu',
    'time': 'masa',
    'tk': 'tak',
    'tnggongjwb': 'tanggungjawab',
    'tngok': 'tengok',
    'tngu': 'tunggu',
    'to': 'kepada',
    'tosak': 'rosak',
    'tp': 'tapi',
    'tpi': 'tapi',
    'tpon': 'telefon',
    'transfer': 'pindah',
    'trgelak': 'tergelak',
    'ts': 'tan sri',
    'tstony': 'tan sri tony',
    'tu': 'itu',
    'tuh': 'itu',
    'tula': 'itulah',
    'umeno': 'umno',
    'unfortunately': 'malangnya',
    'unhappy': 'tidak gembira',
    'up': 'naik',
    'upkan': 'naikkan',
    'ur': 'awak',
    'utk': 'untuk',
    'very': 'sangat',
    'viral': 'tular',
    'vote': 'undi',
    'warning': 'amaran',
    'warranty': 'waranti',
    'wassap': 'whatsapp',
    'wat': 'apa',
    'weii': 'wei',
    'well': 'maklumlah',
    'win': 'menang',
    'with': 'dengan',
    'wt': 'buat',
    'x': 'tak',
    'tw': 'tahu',
    'ye': 'ya',
    'yee': 'ya',
    'yg': 'yang',
    'yng': 'yang',
    'you': 'awak',
    'your': 'awak',
    'sakai': 'selekeh',
    'rmb': 'billion ringgit',
    'rmj': 'juta ringgit',
    'rmk': 'ribu ringgit',
    'rm': 'ringgit',
}



import malaya
import pandas as pd

df = pd.read_csv('tests/02032018.csv',sep=';')
df = df.iloc[3:,1:]
df.columns = ['text','label']
corpus = df.text.tolist()

def test_lda():
    lda = malaya.lda_topic_modelling(corpus,10)
    assert len(lda.get_topics(10))

def test_lda_print():
    lda = malaya.lda_topic_modelling(corpus,10)
    lda.print_topics(5)

def test_lda_get_sentences():
    lda = malaya.lda_topic_modelling(corpus,10)
    assert len(lda.get_sentences(5))

def test_nmf():
    nmf = malaya.nmf_topic_modelling(corpus,10)
    assert len(nmf.get_topics(10))

def test_lsa():
    lsa = malaya.lsa_topic_modelling(corpus,10)
    assert len(lsa.get_topics(10))



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from ._utils import _tag_class
from ._utils._paths import PATH_ENTITIES, S3_PATH_ENTITIES


def available_deep_model():
    """
    List available deep learning entities models, ['concat', 'bahdanau', 'luong', 'entity-network', 'attention']
    """
    return ['concat', 'bahdanau', 'luong', 'entity-network', 'attention']


def crf(validate = True):
    """
    Load CRF Entities Recognition model.

    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    CRF : malaya._models._sklearn_model.CRF class
    """
    return _tag_class.crf(
        PATH_ENTITIES, S3_PATH_ENTITIES, 'entity', validate = validate
    )


def deep_model(model = 'bahdanau', validate = True):
    """
    Load deep learning NER model.

    Parameters
    ----------
    model : str, optional (default='bahdanau')
        Model architecture supported. Allowed values:

        * ``'concat'`` - Concating character and word embedded for BiLSTM
        * ``'bahdanau'`` - Concating character and word embedded including Bahdanau Attention for BiLSTM
        * ``'luong'`` - Concating character and word embedded including Luong Attention for BiLSTM
        * ``'entity-network'`` - Concating character and word embedded on hybrid Entity-Network and RNN
        * ``'attention'`` - Concating character and word embedded with self-attention for BiLSTM
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    TAGGING: malaya._models._tensorflow_model.TAGGING class
    """
    return _tag_class.deep_model(
        PATH_ENTITIES,
        S3_PATH_ENTITIES,
        'entity',
        model = model,
        validate = validate,
    )



import malaya

def test_lang_labels():
    assert malaya.get_language_labels()[0] == 'OTHER'

def test_multinomial_lang_sentence():
    multinomial = malaya.multinomial_detect_languages()
    malay_text = 'beliau berkata program Inisitif Peduli Rakyat (IPR) yang diperkenalkan oleh kerajaan negeri Selangor lebih besar sumbangannya'
    assert multinomial.predict(malay_text) == 'MALAY'

def test_multinomial_lang_sentence_proba():
    multinomial = malaya.multinomial_detect_languages()
    malay_text = 'beliau berkata program Inisitif Peduli Rakyat (IPR) yang diperkenalkan oleh kerajaan negeri Selangor lebih besar sumbangannya'
    assert multinomial.predict(malay_text,get_proba=True)['MALAY'] > 0

def test_multinomial_lang_sentences():
    multinomial = malaya.multinomial_detect_languages()
    malay_text = 'beliau berkata program Inisitif Peduli Rakyat (IPR) yang diperkenalkan oleh kerajaan negeri Selangor lebih besar sumbangannya'
    assert multinomial.predict_batch([malay_text,malay_text])[0] == 'MALAY'

def test_multinomial_lang_sentences_proba():
    multinomial = malaya.multinomial_detect_languages()
    malay_text = 'beliau berkata program Inisitif Peduli Rakyat (IPR) yang diperkenalkan oleh kerajaan negeri Selangor lebih besar sumbangannya'
    assert multinomial.predict_batch([malay_text,malay_text],get_proba=True)[0]['MALAY'] > 0

def test_xgb_lang_sentence():
    xgb = malaya.xgb_detect_languages()
    malay_text = 'beliau berkata program Inisitif Peduli Rakyat (IPR) yang diperkenalkan oleh kerajaan negeri Selangor lebih besar sumbangannya'
    assert xgb.predict(malay_text) == 'MALAY'

def test_xgb_lang_sentence_proba():
    xgb = malaya.xgb_detect_languages()
    malay_text = 'beliau berkata program Inisitif Peduli Rakyat (IPR) yang diperkenalkan oleh kerajaan negeri Selangor lebih besar sumbangannya'
    assert xgb.predict(malay_text,get_proba=True)['MALAY'] > 0

def test_xgb_lang_sentences():
    xgb = malaya.xgb_detect_languages()
    malay_text = 'beliau berkata program Inisitif Peduli Rakyat (IPR) yang diperkenalkan oleh kerajaan negeri Selangor lebih besar sumbangannya'
    assert xgb.predict_batch([malay_text,malay_text])[0] == 'MALAY'

def test_xgb_lang_sentences_proba():
    xgb = malaya.xgb_detect_languages()
    malay_text = 'beliau berkata program Inisitif Peduli Rakyat (IPR) yang diperkenalkan oleh kerajaan negeri Selangor lebih besar sumbangannya'
    assert xgb.predict_batch([malay_text,malay_text],get_proba=True)[0]['MALAY'] > 0



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import pickle
import json
import tensorflow as tf
from ._utils import check_file, load_graph, check_available
from ..texts._text_functions import entities_textcleaning
from .._models._tensorflow_model import TAGGING
from .._models._sklearn_model import CRF


def crf(path, s3_path, class_name, is_lower = True, validate = True):
    if validate:
        check_file(path['crf'], s3_path['crf'])
    else:
        if not check_available(path['crf']):
            raise Exception(
                '%s/crf is not available, please `validate = True`'
                % (class_name)
            )
    try:
        with open(path['crf']['model'], 'rb') as fopen:
            model = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('%s/crf') and try again"
            % (class_name)
        )
    return CRF(model, is_lower = is_lower)


def deep_model(
    path,
    s3_path,
    class_name,
    model = 'bahdanau',
    is_lower = True,
    validate = True,
):
    """
    Load deep learning NER model.

    Parameters
    ----------
    model : str, optional (default='bahdanau')
        Model architecture supported. Allowed values:

        * ``'concat'`` - Concating character and word embedded for BiLSTM
        * ``'bahdanau'`` - Concating character and word embedded including Bahdanau Attention for BiLSTM
        * ``'luong'`` - Concating character and word embedded including Luong Attention for BiLSTM
        * ``'entity-network'`` - Concating character and word embedded on hybrid Entity-Network and RNN
        * ``'attention'`` - Concating character and word embedded with self-attention for BiLSTM

    Returns
    -------
    TAGGING: malaya._models._tensorflow_model.TAGGING class
    """
    assert isinstance(model, str), 'model must be a string'
    model = model.lower()
    if model in ['concat', 'bahdanau', 'luong', 'entity-network', 'attention']:
        if validate:
            check_file(path[model], s3_path[model])
        else:
            if not check_available(path[model]):
                raise Exception(
                    '%s/%s is not available, please `validate = True`'
                    % (class_name, model)
                )
        try:
            with open(path[model]['setting'], 'r') as fopen:
                nodes = json.loads(fopen.read())
            g = load_graph(path[model]['model'])
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('%s/%s') and try again"
                % (class_name, model)
            )
        if model == 'entity-network':
            return TAGGING(
                g.get_tensor_by_name('import/question:0'),
                g.get_tensor_by_name('import/char_ids:0'),
                g.get_tensor_by_name('import/logits:0'),
                nodes,
                tf.InteractiveSession(graph = g),
                model,
                g.get_tensor_by_name('import/transitions:0'),
                g.get_tensor_by_name('import/Variable:0'),
                is_lower = is_lower,
                story = g.get_tensor_by_name('import/story:0'),
            )
        else:
            return TAGGING(
                g.get_tensor_by_name('import/Placeholder:0'),
                g.get_tensor_by_name('import/Placeholder_1:0'),
                g.get_tensor_by_name('import/logits:0'),
                nodes,
                tf.InteractiveSession(graph = g),
                model,
                g.get_tensor_by_name('import/transitions:0'),
                g.get_tensor_by_name('import/Variable:0'),
                is_lower = is_lower,
            )

    else:
        raise Exception(
            'model not supported, please check supported models from malaya.%s.available_deep_model()'
            % (class_name)
        )



# Malaya Natural Language Toolkit
#
# Copyright (C) 2019 Malaya Project
# Licensed under the MIT License
# Author: huseinzol05 <husein.zol05@gmail.com>
# URL: <https://malaya.readthedocs.io/>
# For license information, see https://github.com/huseinzol05/Malaya/blob/master/LICENSE

import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import os
from shutil import rmtree
from pathlib import Path

home = os.path.join(str(Path.home()), 'Malaya')
version = '1.5'
bump_version = '1.5.0.0'
version_path = os.path.join(home, 'version')


def _delete_folder(folder):
    for root, dirs, files in os.walk(folder):
        for file in files:
            os.remove(os.path.join(root, file))


def _delete_macos():
    macos = os.path.join(home, '__MACOSX')
    if os.path.exists(macos):
        rmtree(macos)


from ._utils._paths import MALAY_TEXT
from ._utils._utils import DisplayablePath, download_file

try:
    if not os.path.exists(home):
        os.makedirs(home)
except:
    raise Exception(
        'Malaya cannot make directory for caching. Please check your '
        + str(Path.home())
    )

_delete_macos()

if not os.path.isfile(version_path):
    print('not found any version, deleting previous version models..')
    _delete_folder(home)
    with open(version_path, 'w') as fopen:
        fopen.write(version)
else:
    with open(version_path, 'r') as fopen:
        cached_version = fopen.read()
    try:
        if float(cached_version) < 1:
            print(
                'Found old version of Malaya, deleting previous version models..'
            )
            _delete_folder(home)
            with open(version_path, 'w') as fopen:
                fopen.write(version)
    except:
        print('Found old version of Malaya, deleting previous version models..')
        _delete_folder(home)
        with open(version_path, 'w') as fopen:
            fopen.write(version)


def print_cache():
    paths = DisplayablePath.make_tree(Path(home))
    for path in paths:
        print(path.displayable())


def clear_all_cache():
    """
    Remove cached data, this will delete entire cache folder.
    """
    _delete_macos()
    try:
        print('clearing cached models..')
        _delete_folder(home)
        with open(version_path, 'w') as fopen:
            fopen.write(version)
        return True
    except:
        print(
            'failed to clear cached models. Please make sure %s is able to overwrite from Malaya'
            % (home)
        )


def clear_cache(location):
    """
    Remove selected cached data, please run malaya.print_cache() to get path.
    """
    assert isinstance(location, str), 'location must be a string'
    location = os.path.join(home, location)
    if not os.path.exists(location):
        raise Exception(
            'folder not exist, please check path from malaya.print_cache()'
        )
    if not os.path.isdir(location):
        raise Exception(
            'Please use parent directory, please check path from malaya.print_cache()'
        )
    _delete_folder(location)
    return True


def load_malay_dictionary():
    """
    load Pustaka dictionary for Spelling Corrector or anything.

    Returns
    -------
    list: list of strings
    """
    if not os.path.isfile(MALAY_TEXT):
        print('downloading Malay texts')
        download_file('v6/malay-text.txt', MALAY_TEXT)
    try:
        with open(MALAY_TEXT, 'r') as fopen:
            results = [
                text.lower()
                for text in (list(filter(None, fopen.read().split('\n'))))
            ]
            if len(results) < 20000:
                raise Exception(
                    "model corrupted due to some reasons, please run malaya.clear_cache('dictionary') and try again"
                )
            return results
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('dictionary') and try again"
        )


def describe_pos_malaya():
    """
    Describe Malaya Part-Of-Speech supported (deprecated, use describe_pos() instead).
    """
    print(
        'This classes are deprecated, we prefer to use `malaya.describe_pos()`'
    )
    print('KT - Kata Tanya')
    print('KJ - Kata Kerja')
    print('KP - Kata Perintah')
    print('KPA - Kata Pangkal')
    print('KB - Kata Bantu')
    print('KPENGUAT - Kata Penguat')
    print('KPENEGAS - Kata Penegas')
    print('NAFI - Kata Nafi')
    print('KPEMERI - Kata Pemeri')
    print('KS - Kata Sendi')
    print('KPEMBENAR - Kata Pembenar')
    print('NAFI - Kata Nafi')
    print('NO - Numbers')
    print('SUKU - Suku Bilangan')
    print('PISAHAN - Kata Pisahan')
    print('KETERANGAN - Kata Keterangan')
    print('ARAH - Kata Arah')
    print('KH - Kata Hubung')
    print('GN - Ganti Nama')
    print('KA - Kata Adjektif')
    print('O - not related, out scope')


def describe_pos():
    """
    Describe Part-Of-Speech supported.
    """
    print('ADJ - Adjective, kata sifat')
    print('ADP - Adposition')
    print('ADV - Adverb, kata keterangan')
    print('ADX - Auxiliary verb, kata kerja tambahan')
    print('CCONJ - Coordinating conjuction, kata hubung')
    print('DET - Determiner, kata penentu')
    print('NOUN - Noun, kata nama')
    print('NUM - Number, nombor')
    print('PART - Particle')
    print('PRON - Pronoun, kata ganti')
    print('PROPN - Proper noun, kata ganti nama khas')
    print('SCONJ - Subordinating conjunction')
    print('SYM - Symbol')
    print('VERB - Verb, kata kerja')
    print('X - Other')


def describe_entities_malaya():
    """
    Describe Malaya Entities supported (deprecated, use describe_entities() instead).
    """
    print(
        'This classes are deprecated, we prefer to use `malaya.describe_entities()`'
    )
    print('PRN - person, group of people, believes, etc')
    print('LOC - location')
    print('NORP - Military, police, government, Parties, etc')
    print('ORG - Organization, company')
    print('LAW - related law document, etc')
    print('ART - art of work, special names, etc')
    print('EVENT - event happen, etc')
    print('FAC - facility, hospitals, clinics, etc')
    print('TIME - date, day, time, etc')
    print('O - not related, out scope')


def describe_entities():
    """
    Describe Entities supported.
    """
    print('OTHER - Other')
    print('law - law, regulation, related law documents, documents, etc')
    print('location - location, place')
    print('organization - organization, company, government, facilities, etc')
    print('person - person, group of people, believes, etc')
    print('quantity - numbers, quantity')
    print('time - date, day, time, etc')
    print('event - unique event happened, etc')


def describe_dependency():
    """
    Describe Dependency supported.
    """
    print('acl - clausal modifier of noun')
    print('advcl - adverbial clause modifier')
    print('advmod - adverbial modifier')
    print('amod - adjectival modifier')
    print('appos - appositional modifier')
    print('aux - auxiliary')
    print('case - case marking')
    print('ccomp - clausal complement')
    print('compound - compound')
    print('compound:plur - plural compound')
    print('conj - conjunct')
    print('cop - cop')
    print('csubj - clausal subject')
    print('dep - dependent')
    print('det - determiner')
    print('fixed - multi-word expression')
    print('flat - name')
    print('iobj - indirect object')
    print('mark - marker')
    print('nmod - nominal modifier')
    print('nsubj - nominal subject')
    print('obj - direct object')
    print('parataxis - parataxis')
    print('root - root')
    print('xcomp - open clausal complement')
    print(
        'you can read more from https://universaldependencies.org/en/dep/xcomp.html'
    )


from . import cluster
from . import dependency
from . import emotion
from . import entity
from . import fast_text
from . import language_detection
from . import normalize
from . import num2word
from . import pos
from . import sentiment
from . import spell
from . import stack
from . import stem
from . import subjective
from . import summarize
from . import topic_influencer
from . import topic_model
from . import toxic
from . import word_mover
from . import word2vec
from .texts import vectorizer



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import numpy as np
import collections
import pandas as pd
from sklearn.utils import shuffle
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.decomposition import TruncatedSVD, NMF, LatentDirichletAllocation
from .stem import sastrawi
from ._models._lda2vec import LDA2VEC
from .texts._text_functions import (
    simple_textcleaning,
    STOPWORDS,
    classification_textcleaning,
    print_topics_modelling,
    build_dataset,
)
from .texts.vectorizer import skipgrams


def _softmax(x):
    e_x = np.exp(x - np.max(x))
    out = e_x / e_x.sum()
    return out


def _softmax_2d(x):
    y = x - x.max(axis = 1, keepdims = True)
    np.exp(y, out = y)
    y /= y.sum(axis = 1, keepdims = True)
    return y


def _prob_words(context, vocab, temperature = 1.0):
    """ This calculates a softmax over the vocabulary as a function
    of the dot product of context and word.
    """
    dot = np.dot(vocab, context)
    prob = _softmax(dot / temperature)
    return prob


def _prepare_topics(
    weights,
    factors,
    word_vectors,
    vocab,
    temperature = 1.0,
    doc_lengths = None,
    term_frequency = None,
    normalize = False,
):
    topic_to_word = []
    msg = 'Vocabulary size did not match size of word vectors'
    assert len(vocab) == word_vectors.shape[0], msg
    if normalize:
        word_vectors /= np.linalg.norm(word_vectors, axis = 1)[:, None]
    for factor_vector in factors:
        factor_to_word = _prob_words(
            factor_vector, word_vectors, temperature = temperature
        )
        topic_to_word.append(np.ravel(factor_to_word))
    topic_to_word = np.array(topic_to_word)
    msg = 'Not all rows in topic_to_word sum to 1'
    assert np.allclose(np.sum(topic_to_word, axis = 1), 1), msg
    doc_to_topic = _softmax_2d(weights)
    msg = 'Not all rows in doc_to_topic sum to 1'
    assert np.allclose(np.sum(doc_to_topic, axis = 1), 1), msg
    data = {
        'topic_term_dists': topic_to_word,
        'doc_topic_dists': doc_to_topic,
        'doc_lengths': doc_lengths,
        'vocab': vocab,
        'term_frequency': term_frequency,
    }
    return data


class _DEEP_TOPIC:
    def __init__(
        self,
        model,
        dictionary,
        reversed_dictionary,
        freqs,
        len_idx_text_clean,
        corpus,
    ):
        self._model = model
        self._dictionary = dictionary
        self._reversed_dictionary = reversed_dictionary
        self._doc_embed = model.sess.run(model.doc_embedding)
        self._topic_embed = model.sess.run(model.topic_embedding)
        self._word_embed = model.sess.run(model.word_embedding)
        self._components = self._topic_embed.dot(self._word_embed.T)
        self._freqs = freqs
        self._doc_len = len_idx_text_clean
        self._corpus = corpus
        self._features = []
        for i in range(len(self._dictionary)):
            self._features.append(self._reversed_dictionary[i])

    def visualize_topics(self, notebook_mode = False, mds = 'pcoa'):
        """
        Print important topics based on decomposition.

        Parameters
        ----------
        mds : str, optional (default='pcoa')
            2D Decomposition. Allowed values:

            * ``'pcoa'`` - Dimension reduction via Jensen-Shannon Divergence & Principal Coordinate Analysis (aka Classical Multidimensional Scaling)
            * ``'mmds'`` - Dimension reduction via Multidimensional scaling
            * ``'tsne'`` - Dimension reduction via t-distributed stochastic neighbor embedding
        """
        assert isinstance(mds, str), 'mds must be a string'
        assert isinstance(
            notebook_mode, bool
        ), 'notebook_mode must be a boolean'
        try:
            import pyLDAvis
            import pyLDAvis.sklearn
        except:
            raise Exception(
                'pyldavis not installed. Please install it and try again.'
            )

        if notebook_mode:
            pyLDAvis.enable_notebook()

        vis_data = _prepare_topics(
            self._doc_embed,
            self._topic_embed,
            self._word_embed,
            np.array(self._features),
            doc_lengths = self._doc_len,
            term_frequency = self._freqs,
            normalize = True,
        )
        prepared_vis_data = pyLDAvis.prepare(**vis_data)
        if notebook_mode:
            return prepared_vis_data
        else:
            pyLDAvis.show(prepared_vis_data)

    def print_topics(self, len_topic):
        """
        Print important topics based on decomposition.

        Parameters
        ----------
        len_topic: int
        """
        assert isinstance(len_topic, int), 'len_topic must be an integer'
        print_topics_modelling(
            range(len_topic),
            feature_names = np.array(self._features),
            sorting = np.argsort(self._components)[:, ::-1],
            topics_per_chunk = 5,
            n_words = 10,
        )

    def get_topics(self, len_topic):
        """
        Return important topics based on decomposition.

        Parameters
        ----------
        len_topic: int

        Returns
        -------
        results: list of strings
        """
        assert isinstance(len_topic, int), 'len_topic must be an integer'
        results = []
        for no, topic in enumerate(self._components):
            results.append(
                (
                    no,
                    ' '.join(
                        [
                            self._features[i]
                            for i in topic.argsort()[: -len_topic - 1 : -1]
                        ]
                    ),
                )
            )
        return results

    def get_sentences(self, len_sentence, k = 0):
        """
        Return important sentences related to selected column based on decomposition.

        Parameters
        ----------
        len_sentence: int
        k: int, (default=0)
            index of decomposition matrix.

        Returns
        -------
        results: list of strings
        """
        assert isinstance(len_sentence, int), 'len_sentence must be an integer'
        assert isinstance(k, int), 'k must be an integer'
        assert (
            k < self._doc_embed.shape[1] and k >= 0
        ), 'k should be between 0 and n_topics'
        reverse_sorted = np.argsort(self._doc_embed[:, k])[::-1]
        return [self._corpus[i] for i in reverse_sorted[:len_sentence]]


class _TOPIC:
    def __init__(
        self, features, comp, corpus, transformed, vectorizer, vectors
    ):
        self.features = features
        self.comp = comp
        self.corpus = corpus
        self.transformed = transformed
        self.vectorizer = vectorizer
        self._vectors = vectors

    def visualize_topics(self, notebook_mode = False, mds = 'pcoa'):
        """
        Print important topics based on decomposition.

        Parameters
        ----------
        mds : str, optional (default='pcoa')
            2D Decomposition. Allowed values:

            * ``'pcoa'`` - Dimension reduction via Jensen-Shannon Divergence & Principal Coordinate Analysis (aka Classical Multidimensional Scaling)
            * ``'mmds'`` - Dimension reduction via Multidimensional scaling
            * ``'tsne'`` - Dimension reduction via t-distributed stochastic neighbor embedding
        """
        assert isinstance(mds, str), 'mds must be a string'
        assert isinstance(
            notebook_mode, bool
        ), 'notebook_mode must be a boolean'
        assert isinstance(
            self.comp, LatentDirichletAllocation
        ), 'only support lda_topic_modelling()'
        import pyLDAvis
        import pyLDAvis.sklearn

        if notebook_mode:
            pyLDAvis.enable_notebook()

        prepared_vis_data = pyLDAvis.sklearn.prepare(
            self.comp, self._vectors, self.vectorizer, mds = mds
        )
        if notebook_mode:
            return prepared_vis_data
        else:
            pyLDAvis.show(prepared_vis_data)

    def print_topics(self, len_topic):
        """
        Print important topics based on decomposition.

        Parameters
        ----------
        len_topic: int
        """
        assert isinstance(len_topic, int), 'len_topic must be an integer'
        print_topics_modelling(
            range(len_topic),
            feature_names = np.array(self.features),
            sorting = np.argsort(self.comp.components_)[:, ::-1],
            topics_per_chunk = 5,
            n_words = 10,
        )

    def get_topics(self, len_topic):
        """
        Return important topics based on decomposition.

        Parameters
        ----------
        len_topic: int

        Returns
        -------
        results: list of strings
        """
        assert isinstance(len_topic, int), 'len_topic must be an integer'
        results = []
        for no, topic in enumerate(self.comp.components_):
            results.append(
                (
                    no,
                    ' '.join(
                        [
                            self.features[i]
                            for i in topic.argsort()[: -len_topic - 1 : -1]
                        ]
                    ),
                )
            )
        return results

    def get_sentences(self, len_sentence, k = 0):
        """
        Return important sentences related to selected column based on decomposition.

        Parameters
        ----------
        len_sentence: int
        k: int, (default=0)
            index of decomposition matrix.

        Returns
        -------
        results: list of strings
        """
        assert isinstance(len_sentence, int), 'len_sentence must be an integer'
        assert isinstance(k, int), 'k must be an integer'
        assert (
            k < self.transformed.shape[1] and k >= 0
        ), 'k should be between 0 and n_topics'
        reverse_sorted = np.argsort(self.transformed[:, k])[::-1]
        return [self.corpus[i] for i in reverse_sorted[:len_sentence]]


def _base_topic_modelling(
    corpus,
    n_topics,
    decomposition,
    max_df = 0.95,
    min_df = 2,
    vectorizer = 'bow',
    stemming = True,
    cleaning = simple_textcleaning,
    stop_words = STOPWORDS,
    **kwargs
):
    assert isinstance(corpus, list) and isinstance(
        corpus[0], str
    ), 'input must be list of strings'
    assert isinstance(n_topics, int), 'n_topics must be an integer'
    assert isinstance(max_df, float), 'max_df must be a float'
    assert isinstance(min_df, int), 'min_df must be an integer'
    assert isinstance(vectorizer, str), 'vectorizer must be a string'
    vectorizer = vectorizer.lower()
    assert vectorizer in ['tfidf', 'bow'], "vectorizer must be 'tfidf' or 'bow'"
    assert (
        max_df < 1 and max_df > 0
    ), 'max_df must be bigger than 0, less than 1'
    assert min_df > 0, 'min_df must be bigger than 0'
    assert isinstance(stemming, bool), 'bool must be a boolean'
    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi(corpus[i])
    if vectorizer == 'tfidf':
        Vectorizer = TfidfVectorizer
    elif vectorizer == 'bow':
        Vectorizer = CountVectorizer
    else:
        raise Exception(
            "vectorizer not support, vectorizer must be 'tfidf' or 'bow'"
        )
    tf_vectorizer = Vectorizer(
        max_df = max_df, min_df = min_df, stop_words = stop_words, **kwargs
    )
    tf = tf_vectorizer.fit_transform(corpus)
    tf_features = tf_vectorizer.get_feature_names()
    compose = decomposition(n_topics).fit(tf)
    return _TOPIC(
        tf_features,
        compose,
        [classification_textcleaning(c) for c in corpus],
        compose.transform(tf),
        tf_vectorizer,
        tf,
    )


def _lda2vec_preprocessing(
    corpus,
    window_size = 2,
    stemming = True,
    cleaning = simple_textcleaning,
    stop_words = STOPWORDS,
):

    assert isinstance(corpus, list) and isinstance(
        corpus[0], str
    ), 'input must be list of strings'
    assert isinstance(stemming, bool), 'bool must be a boolean'
    assert isinstance(window_size, int), 'window_size must be an integer'
    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi(corpus[i])
    text_clean = []
    for text in corpus:
        text_clean.append(
            ' '.join([word for word in text.split() if word not in stop_words])
        )
    concat = (' '.join(text_clean)).split()
    unique_count = len(list(set(concat)))
    _, count, dictionary, reversed_dictionary = build_dataset(
        concat, unique_count, included_prefix = False
    )
    idx_text_clean, len_idx_text_clean = [], []
    for text in text_clean:
        splitted = [dictionary[word] for word in text.split()]
        idx_text_clean.append(splitted)
        len_idx_text_clean.append(len(splitted))
    word_counts = collections.Counter(concat)
    df_freqs = pd.DataFrame.from_records(word_counts, index = ['Freqs'])
    df_freqs = df_freqs.T
    df_freqs = df_freqs.sort_values(['Freqs'], ascending = False)
    freqs = df_freqs.values.flatten().tolist()[:unique_count]
    doc_ids = np.arange(len(idx_text_clean))
    num_unique_documents = doc_ids.max()
    skipgrams_data = []
    for i, t in enumerate(idx_text_clean):
        pairs, _ = skipgrams(
            t,
            vocabulary_size = unique_count,
            window_size = window_size,
            shuffle = True,
            negative_samples = 0,
        )
        for pair in pairs:
            temp_data = pair
            temp_data.append(doc_ids[i])
            skipgrams_data.append(temp_data)
    skipgrams_data_df = pd.DataFrame(skipgrams_data)
    pivot_words = skipgrams_data_df[0].values
    target_words = skipgrams_data_df[1].values
    doc_ids = skipgrams_data_df[2].values
    pivot_words, target_words, doc_ids = shuffle(
        pivot_words, target_words, doc_ids, random_state = 10
    )
    num_unique_documents = doc_ids.max() + 1
    return (
        pivot_words,
        target_words,
        doc_ids,
        dictionary,
        reversed_dictionary,
        num_unique_documents,
        freqs,
        len_idx_text_clean,
        corpus,
    )


def lda(
    corpus,
    n_topics = 10,
    max_df = 0.95,
    min_df = 2,
    stemming = True,
    vectorizer = 'bow',
    cleaning = simple_textcleaning,
    stop_words = STOPWORDS,
    **kwargs
):
    """
    Train a LDA model to do topic modelling based on corpus / list of strings given.

    Parameters
    ----------
    corpus: list
    n_topics: int, (default=10)
        size of decomposition column.
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    stemming: bool, (default=True)
        If True, sastrawi_stemmer will apply
    vectorizer: str, (default='bow')
        vectorization technique for corpus, only support 'bow' and 'tfidf'.
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus
    stop_words: list, (default=STOPWORDS)
        list of stop words to remove

    Returns
    -------
    _TOPIC: malaya.topic_modelling._TOPIC class
    """
    return _base_topic_modelling(
        corpus,
        n_topics,
        LatentDirichletAllocation,
        max_df = max_df,
        min_df = min_df,
        vectorizer = vectorizer,
        stemming = stemming,
        cleaning = cleaning,
        stop_words = stop_words,
        **kwargs
    )


def nmf(
    corpus,
    n_topics = 10,
    max_df = 0.95,
    min_df = 2,
    stemming = True,
    vectorizer = 'bow',
    cleaning = simple_textcleaning,
    stop_words = STOPWORDS,
    **kwargs
):
    """
    Train a NMF model to do topic modelling based on corpus / list of strings given.

    Parameters
    ----------
    corpus: list
    n_topics: int, (default=10)
        size of decomposition column.
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    stemming: bool, (default=True)
        If True, sastrawi_stemmer will apply.
    vectorizer: str, (default='bow')
        vectorization technique for corpus, only support 'bow' and 'tfidf'.
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus.
    stop_words: list, (default=STOPWORDS)
        list of stop words to remove.

    Returns
    -------
    _TOPIC: malaya.topic_modelling._TOPIC class
    """
    return _base_topic_modelling(
        corpus,
        n_topics,
        NMF,
        max_df = max_df,
        min_df = min_df,
        vectorizer = vectorizer,
        stemming = stemming,
        cleaning = cleaning,
        stop_words = stop_words,
        **kwargs
    )


def lsa(
    corpus,
    n_topics,
    max_df = 0.95,
    min_df = 2,
    vectorizer = 'bow',
    stemming = True,
    cleaning = simple_textcleaning,
    stop_words = STOPWORDS,
    **kwargs
):
    """
    Train a LSA model to do topic modelling based on corpus / list of strings given.

    Parameters
    ----------
    corpus: list
    n_topics: int, (default=10)
        size of decomposition column.
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    vectorizer: str, (default='bow')
        vectorization technique for corpus, only support 'bow' and 'tfidf'.
    stemming: bool, (default=True)
        If True, sastrawi_stemmer will apply
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus
    stop_words: list, (default=STOPWORDS)
        list of stop words to remove

    Returns
    -------
    _TOPIC: malaya.topic_modelling._TOPIC class
    """
    return _base_topic_modelling(
        corpus,
        n_topics,
        TruncatedSVD,
        max_df = max_df,
        min_df = min_df,
        vectorizer = vectorizer,
        stemming = stemming,
        cleaning = cleaning,
        stop_words = stop_words,
        **kwargs
    )


def lda2vec(
    corpus,
    n_topics,
    stemming = True,
    cleaning = simple_textcleaning,
    stop_words = STOPWORDS,
    window_size = 2,
    embedding_size = 128,
    training_iteration = 10,
    switch_loss = 3,
    **kwargs
):
    """
    Train a LDA2Vec model to do topic modelling based on corpus / list of strings given.

    Parameters
    ----------
    corpus: list
    n_topics: int, (default=10)
        size of decomposition column.
    stemming: bool, (default=True)
        If True, sastrawi_stemmer will apply
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus
    stop_words: list, (default=STOPWORDS)
        list of stop words to remove
    embedding_size: int, (default=128)
        embedding size of lda2vec tensors
    training_iteration: int, (default=10)
        training iteration, how many loop need to train
    switch_loss: int, (default=3)
        baseline to switch from document based loss to document + word based loss

    Returns
    -------
    _DEEP_TOPIC: malaya.topic_modelling._DEEP_TOPIC class
    """
    pivot_words, target_words, doc_ids, dictionary, reversed_dictionary, num_unique_documents, freqs, len_idx_text_clean, cleaned_corpus = _lda2vec_preprocessing(
        corpus,
        window_size = window_size,
        stemming = stemming,
        cleaning = cleaning,
        stop_words = stop_words,
    )
    model = LDA2VEC(
        num_unique_documents,
        len(dictionary),
        n_topics,
        freqs,
        embedding_size = embedding_size,
        **kwargs
    )
    model.train(
        pivot_words,
        target_words,
        doc_ids,
        training_iteration,
        switch_loss = switch_loss,
    )
    return _DEEP_TOPIC(
        model,
        dictionary,
        reversed_dictionary,
        freqs,
        len_idx_text_clean,
        cleaned_corpus,
    )



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from collections import Counter, defaultdict
from fuzzywuzzy import fuzz
import numpy as np
from .texts._text_functions import normalizer_textcleaning
from .topic_influencer import is_location
from .texts._tatabahasa import alphabet, consonants, vowels


def _build_dicts(words):
    occurences = {}
    for l in alphabet:
        occurences[l] = defaultdict(lambda: 0)
    for word in words:
        occurences[word[0]][word] += 1
    return occurences


def _edit_normalizer(word):
    splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
    deletes = [a + b[1:] for a, b in splits if b]
    transposes = [a + b[1] + b[0] + b[2:] for a, b in splits if len(b) > 1]
    replaces = [a + c + b[1:] for a, b in splits for c in alphabet if b]
    inserts = [a + c + b for a, b in splits for c in alphabet]
    pseudo = []
    for c in vowels:
        pseudo.append(''.join([w + c for w in word]))
    fuzziness = []
    if len(word):
        if word[-1] == 'e':
            fuzziness.append(word[:-1] + 'a')
    if len(word):
        if word[-2:] == 'ak':
            fuzziness.append(word[:-1])
    if len(word) > 2:
        if (
            word[0] in consonants
            and word[1] in consonants
            and word[2] in vowels
        ):
            fuzziness.append(word[0] + word[2] + word[1:])
    if len(word) > 2:
        if (
            word[0] in vowels
            and word[-1] in consonants
            and word[2] in consonants
        ):
            fuzziness.append(word[:-2] + word[0] + word[-1])
    if len(word) > 2:
        if word[-1] == 'o' and word[-3] in vowels and word[-2] in consonants:
            fuzziness.append(word[:-1] + 'ar')
    if len(word):
        if word[0] == 'a' and word[1] in consonants:
            fuzziness.append('h' + word)
    return set(deletes + transposes + replaces + inserts + fuzziness + pseudo)


def _return_possible(word, dicts, edits):
    return set(e2 for e1 in edits(word) for e2 in edits(e1) if e2 in dicts)


def _return_known(word, dicts):
    return set(w for w in word if w in dicts)


class _SPELL:
    def __init__(self, corpus):
        self.corpus = corpus
        self.occurences = _build_dicts(self.corpus)
        self.corpus = Counter(corpus)

    def correct(self, string, first_char = True, debug = True):
        """
        Correct a word.

        Parameters
        ----------
        string: str
        first_char: bool, optional (default=True)
            If True, it will only pulled nearest words based on first character, faster but less accurate.
        debug : bool, optional (default=True)
            If true, it will print character similarity distances.

        Returns
        -------
        string: corrected string
        """
        assert (isinstance(string, str)) and not string.count(
            ' '
        ), 'input must be a single word'
        assert len(string) > 1, 'input must be long than 1 characters'
        string = normalizer_textcleaning(string)
        if string.istitle():
            return string
        if not len(string):
            return string
        if first_char:
            candidates = (
                _return_known([string], self.occurences[string[0]])
                or _return_known(
                    _edit_normalizer(string), self.occurences[string[0]]
                )
                or _return_possible(
                    string, self.occurences[string[0]], _edit_normalizer
                )
                or [string]
            )
        else:
            candidates = (
                _return_known([string], self.corpus)
                or _return_known(_edit_normalizer(string), self.corpus)
                or _return_possible(string, self.corpus, _edit_normalizer)
                or [string]
            )
        candidates = [
            (candidate, is_location(candidate))
            for candidate in list(candidates)
        ]
        if debug:
            print([(k, fuzz.ratio(string, k[0])) for k in candidates], '\n')
        strings = [fuzz.ratio(string, k[0]) for k in candidates]
        descending_sort = np.argsort(strings)[::-1]
        for index in descending_sort:
            if not candidates[index][1]:
                return candidates[index][0]
        return candidates[descending_sort[0]][0]


def naive(corpus):
    """
    Train a fuzzy logic Spell Corrector

    Parameters
    ----------
    corpus : list of strings. Prefer to feed with malaya.load_malay_dictionary()

    Returns
    -------
    SPELL: Trained malaya.spell._SPELL class
    """
    assert isinstance(corpus, list) and isinstance(
        corpus[0], str
    ), 'input must be list of strings'
    return _SPELL(corpus)
